import cv2
import numpy as np
import glm
from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype
from .ImageLoader import ImageLoader
from .helper import get_external_format, width_adapt, get_dtype, get_channels

class sampler2DArray(FBOAttachment):

    _default_internal_format = GL.GL_RGBA32F
    _default_filter_min = GL.GL_LINEAR
    _default_filter_mag = GL.GL_LINEAR
    _default_filter_mipmap = GL.GL_LINEAR

    _basic_info = \
    {
        "gen_func": GL.glGenTextures,
        "bind_func": GL.glBindTexture,
        "del_func": GL.glDeleteTextures,
        "target_type": GL.GL_TEXTURE_2D_ARRAY,
        "binding_type": GL.GL_TEXTURE_BINDING_2D_ARRAY,
        "need_number": True,
    }

    class Images:
        def __init__(self, parent):
            self._images = []
            self._parent = parent

        @checktype
        def __getitem__(self, index:int):
            return self._images[index]
        
        @checktype
        def __setitem__(self, index:int, image:(str,np.ndarray)):
            if isinstance(image, str):
                image = ImageLoader.load(image)

            self._images[index] = image

            if image is None:
                return
            
            if image.shape[1] > self._parent.width:
                self._parent.width = image.shape[1]

            if image.shape[0] > self._parent.height:
                self._parent.height = image.shape[0]

            channels = image.shape[2] if len(image.shape) > 2 else 1
            internal_format = GLInfo.internal_formats_map[image.dtype][channels]
            if internal_format != self._parent._internal_format:
                self._parent._internal_format = internal_format
                self._parent._image_size_changed = True

            self._parent._image_layer_changed[index] = True

        def __delitem__(self, index):
            del self._images[index]
            del self._parent._image_layer_changed[index]

        def __contains__(self, image):
            return image in self._images

        def __len__(self):
            return len(self._images)

        def append(self, image:(str,np.ndarray)):
            self._images.append(None)
            index = len(self._images) - 1
            self._parent._image_layer_changed.append(False)
            self._parent._image_size_changed = True

            self[index] = image

        def insert(self, index, image:(str,np.ndarray)):
            self._images.insert(index, image)
            self._parent._image_layer_changed.insert(index, False)
            self._parent._image_size_changed = True

            self[index] = image

        def index(self, image):
            return self._images.index(image)
        
        def clear(self):
            if not self._images:
                return
            
            self._images.clear()
            self._parent._image_layer_changed.clear()
            self._parent._image_size_changed = True

        def pop(self, index:int=None):
            self._images.pop(index)
            self._parent._image_layer_changed.pop(index)
            self._parent._image_size_changed = True
            
    @checktype
    def __init__(self, images:list=None, width:int=None, height:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):
        FBOAttachment.__init__(self)
        
        self._handle = 0
        self._images = sampler2DArray.Images(self)
        self._width = 0
        self._height = 0
        self._internal_format = self.__class__._default_internal_format

        self._wrap_r = GL.GL_CLAMP_TO_EDGE
        self._wrap_s = GL.GL_REPEAT
        self._wrap_t = GL.GL_REPEAT
        self._filter_min = self.__class__._default_filter_min
        self._filter_mag = self.__class__._default_filter_mag
        self._filter_mipmap = self.__class__._default_filter_mipmap
        self._border_color = glm.vec4(0, 0, 0, 1)

        self._image_size_changed = True
        self._image_layer_changed = []
        self._wrap_r_changed = True
        self._wrap_s_changed = True
        self._wrap_t_changed = True
        self._filter_min_changed = True
        self._filter_mag_changed = True
        self._border_color_changed = False

        if images is not None:
            for image in images:
                self._images.append(image)
        elif layers is not None:
            self.layers = layers

        if width is not None:
            self.width = width

        if height is not None:
            self.height = height

        if internal_format is not None:
            self.internal_format = internal_format

    def __del__(self):
        if self._handle != 0:
            # bt.glMakeTextureHandleNonResidentARB(self._handle)
            self._handle = 0

        FBOAttachment.__del__(self)

    def __hash__(self):
        return id(self)

    def __deepcopy__(self, memo):
        result = self.__class__()
        result._images = self._images
        result._width = self._width
        result._height = self._height
        result._internal_format = self._internal_format

        result._wrap_s = self._wrap_s
        result._wrap_t = self._wrap_t
        result._filter_min = self._filter_min
        result._filter_mag = self._filter_mag
        result._filter_mipmap = self._filter_mipmap
        result._border_color = self._border_color

        return result

    @property
    def handle(self):
        if not bt.glGetTextureHandleARB:
            return 0
        
        self.bind()
        if self._handle == 0:
            self._handle = bt.glGetTextureHandleARB(self._id)
            if self._handle == 0:
                raise RuntimeError(f"failed to create sampler2DArray {self._id}'s handle")
            bt.glMakeTextureHandleResidentARB(self._handle)            

        return self._handle

    def bind(self, update_fbo:bool=False, force_update_image:bool=False):
        FBOAttachment.bind(self, update_fbo)

        if self._wrap_r_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_WRAP_R, self._wrap_r)
            self._wrap_r_changed = False
        if self._wrap_s_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_WRAP_S, self._wrap_s)
            self._wrap_s_changed = False
        if self._wrap_t_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_WRAP_T, self._wrap_t)
            self._wrap_t_changed = False
        if self._border_color_changed:
            GL.glTexParameterfv(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_BORDER_COLOR, glm.value_ptr(self._border_color))
            self._border_color_changed = False

        if self._filter_min_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_MIN_FILTER, self._get_filter_min())
            self._filter_min_changed = False
        if self._filter_mag_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D_ARRAY, GL.GL_TEXTURE_MAG_FILTER, self._filter_mag)
            self._filter_mag_changed = False

        generated_mipmap = False
        external_format = get_external_format(self._internal_format)
        if force_update_image or self._image_size_changed:
            width_adapt(self._width)
            GL.glTexImage3D(GL.GL_TEXTURE_2D_ARRAY, 0, self._internal_format,
                            self._width, self._height, self.layers, 0, external_format, self.dtype, None)
            
            if not update_fbo and self._filter_mipmap is not None and self._image is not None:
                self.generate_mipmap()
                generated_mipmap = True

            self._image_size_changed = False

        for layer in range(self.layers):
            if self._image_layer_changed[layer]:
                GL.glTexSubImage3D(GL.GL_TEXTURE_2D_ARRAY, 0, 0, 0, layer, self.width, self.height, 1, external_format, self.dtype, self.images[layer])

        if not update_fbo and not generated_mipmap and not self._fbo_image_generated_mipmap:
            if self._filter_mipmap is not None:
                self.generate_mipmap()

            self._fbo_image_generated_mipmap = True

    def generate_mipmap(self):
        FBOAttachment.bind(self)
        GL.glGenerateMipmap(GL.GL_TEXTURE_2D_ARRAY)

    def clear(self):
        if self._id == 0:
            return

        self.bind()
        external_format = get_external_format(self._internal_format)
        dtype = get_dtype(self._internal_format)
        GL.glClearTexImage(self._id, 0, external_format, dtype, None)

    @property
    def width(self):
        return self._width

    @width.setter
    @FBOAttachment.param_setter
    def width(self, width:int):
        self._width = width
        self._image_size_changed = True

    @property
    def height(self):
        return self._height

    @height.setter
    @FBOAttachment.param_setter
    def height(self, height:int):
        self._height = height
        self._image_size_changed = True

    @property
    def layers(self):
        return len(self._images)

    @layers.setter
    @FBOAttachment.param_setter
    def layers(self, layers:int):
        len_images = len(self._images)
        if layers < len_images:
            del self._images[layers:]

        for i in range(layers - len_images):
            self._images.append(None)

    @property
    def dtype(self):
        return get_dtype(self._internal_format)

    @property
    def internal_format(self):
        return self._internal_format

    def _set_internal_format(self, internal_format):
        self._internal_format = internal_format

        convert_map = \
        {
            (1,3): cv2.COLOR_GRAY2RGB,
            (1,4): cv2.COLOR_GRAY2RGBA,
            (3,1): cv2.COLOR_RGB2GRAY,
            (3,4): cv2.COLOR_RGB2RGBA,
            (4,1): cv2.COLOR_RGBA2GRAY,
            (4,3): cv2.COLOR_RGBA2RGB
        }

        for layer in range(self.layers):
            if self._images[layer] is None:
                continue

            old_channels = self._images[layer].shape[2] if len(self._images[layer].shape) > 2 else 1
            new_channels = get_channels(internal_format)
            if (old_channels, new_channels) in convert_map:
                conversion = convert_map[old_channels, new_channels]
                self._images[layer] = cv2.cvtColor(self._images[layer], conversion)
            elif old_channels == 1:
                if len(self._images[layer].shape) == 2:
                    self._images[layer] = np.expand_dims(self._images[layer], axis=2)
                self._images[layer] = self._images[layer].repeat(new_channels, axis=2)
            elif new_channels > old_channels:
                self._images[layer] = np.append(self._images[layer], np.zeros((self._images[layer].shape[0], self._images[layer].shape[1], new_channels-old_channels), dtype=self._images[layer].dtype), axis=2)
            elif new_channels < old_channels:
                self._images[layer] = self._images[layer][:, :, :new_channels]

            old_np_dtype = self._images[layer].dtype
            new_np_dtype = GLInfo.dtype_map[get_dtype(self._internal_format)]
            if old_np_dtype != new_np_dtype:
                if "int" in str(old_np_dtype) and "float" in str(new_np_dtype):
                    self._images[layer] = (self._images[layer] / 255).astype(new_np_dtype)
                elif "float" in str(old_np_dtype) and "int" in str(new_np_dtype):
                    self._images[layer] = (self._images[layer] * 255).astype(new_np_dtype)
                else:
                    self._images[layer] = self._images[layer].astype(new_np_dtype)

            self._image_layer_changed[layer] = True

    @internal_format.setter
    @FBOAttachment.param_setter
    def internal_format(self, internal_format:GLInfo.internal_formats):
        self._set_internal_format(internal_format)

    @property
    def wrap_r(self):
        return self._wrap_r
    
    @wrap_r.setter
    @FBOAttachment.param_setter
    def wrap_r(self, wrap_type:GLInfo.wrap_types):
        self._wrap_r = wrap_type
        self._wrap_r_changed = True

    @property
    def wrap_s(self):
        return self._wrap_s

    @wrap_s.setter
    @FBOAttachment.param_setter
    def wrap_s(self, wrap_type:GLInfo.wrap_types):
        self._wrap_s = wrap_type
        self._wrap_s_changed = True

    @property
    def wrap_t(self):
        return self._wrap_t

    @wrap_t.setter
    @FBOAttachment.param_setter
    def wrap_t(self, wrap_type:GLInfo.wrap_types):
        self._wrap_t = wrap_type
        self._wrap_t_changed = True

    @property
    def wrap(self):
        return self._wrap_r, self._wrap_s, self._wrap_t

    @wrap.setter
    @FBOAttachment.param_setter
    def wrap(self, wrap_type):
        if wrap_type in GLInfo.wrap_types:
            self.wrap_r = wrap_type
            self.wrap_s = wrap_type
            self.wrap_t = wrap_type
        else:
            self.wrap_r = wrap_type[0]
            self.wrap_s = wrap_type[1]
            self.wrap_t = wrap_type[2]

    @property
    def filter_min(self):
        return self._filter_min

    @filter_min.setter
    @FBOAttachment.param_setter
    def filter_min(self, filter_type:GLInfo.filter_types):
        self._filter_min = filter_type
        self._filter_min_changed = True

    @property
    def filter_mag(self):
        return self._filter_mag

    @filter_mag.setter
    @FBOAttachment.param_setter
    def filter_mag(self, filter_type:GLInfo.filter_types):
        self._filter_mag = filter_type
        self._filter_mag_changed = True

    @property
    def filter_mipmap(self):
        return self._filter_mipmap

    @filter_mipmap.setter
    @FBOAttachment.param_setter
    def filter_mipmap(self, filter_type:GLInfo.filter_types):
        self._filter_mipmap = filter_type
        self._filter_min_changed = True
        self._filter_mag_changed = True

    def _get_filter_min(self):
        if self._filter_mipmap is None:
            return self._filter_min
        else:
            return GLInfo.mipmap_filter_map[(self._filter_min, self._filter_mipmap)]
        
    @property
    def filter(self):
        return self._filter_min, self._filter_mag, self._filter_mipmap

    @filter.setter
    @FBOAttachment.param_setter
    def filter(self, filter_type):
        if filter_type in GLInfo.filter_types:
            self.filter_min = filter_type
            self.filter_mag = filter_type
            self.filter_mipmap = filter_type
        else:
            self.filter_min = filter_type[0]
            self.filter_mag = filter_type[1]
            if len(filter_type) > 2:
                self.filter_mipmap = filter_type[2]

    @property
    def border_color(self):
        return self._border_color

    @border_color.setter
    @FBOAttachment.param_setter
    def border_color(self, color:(glm.vec3, glm.vec4)):
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        self._border_color = color
        self._border_color_changed = True

    @property
    def images(self):
        return self._images

    @checktype
    def malloc(self, width:int, height:int, samples:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):        
        self.width = width
        self.height = height
        
        if layers is not None:
            self.layers = layers

        if internal_format is None:
            internal_format = self.internal_format

        if internal_format is None:
            internal_format = self.__class__._default_internal_format

        self.internal_format = internal_format

    @property
    def is_completed(self):
        return (self._width > 0 and self._height > 0 and self.layers > 0)

    @property
    def samples(self):
        return None
    
    @samples.setter
    def samples(self, samples:int):
        pass



from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
from OpenGL import GL

class Points(Mesh):
    
    def __init__(self, points:list=[], color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), point_size:int=5,
                 name:str="", block:bool=True):
        Mesh.__init__(self, primitive_type=GL.GL_POINTS, color=color, name=name, block=block, shared=False)
        self.render_hints.point_size = point_size
        self.__points = points
        self.start_building()

    def build(self):
        points = self.__points
        vertices = self.vertices

        length = 0
        len_points = len(points)
        for i in range(len_points):
            if i > 0:
                length += glm.length(points[i]-points[i-1])
            vertices[i] = Vertex(position=points[i], tex_coord=glm.vec3(length, 0, 0))

        del vertices[len_points:]

    @property
    def points(self):
        return self.__points
    
    @points.setter
    def points(self, points):
        self.__points = points
        self.start_building()

    @property
    def point_size(self):
        return self.render_hints.point_size
    
    @point_size.setter
    @checktype
    def point_size(self, point_size:int):
        self.render_hints.point_size = point_size


from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class RPolygonFace(Mesh):

    @checktype
    def __init__(self, n_sides:int=5, start_side:int=0, total_sides:int=None, radius:float=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 vertical:bool=False, normalize_tex_coord:bool=False,
                 name:str="", block:bool=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__radius = radius
        self.__start_side = start_side
        self.__total_sides = total_sides
        self.__n_sides = n_sides
        self.__vertical = vertical
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        start_side = self.__start_side
        total_sides = self.__total_sides
        if total_sides is None:
            total_sides = self.__n_sides
        total_sides = min(self.__n_sides, total_sides)
        n_sides = self.__n_sides
        vertical = self.__vertical
        normalize_tex_coord = self.__normalize_tex_coord

        i_vertex = 0
        i_index = 0

        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        # 中心点
        vertex_center = Vertex()
        vertex_center.position = glm.vec3(0)
        vertex_center.normal = normal
        vertex_center.tex_coord = glm.vec3(0.5, 0.5, 0)

        vertices[i_vertex] = vertex_center
        i_vertex += 1

        tex_coord_radius = 0.5 if normalize_tex_coord else radius

        for j in range(total_sides+1):
            theta = 2*math.pi*(j+start_side)/n_sides
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)

            edge = radius*glm.vec3(cos_theta, sin_theta, 0)
            if vertical:
                edge = radius*glm.vec3(cos_theta, 0, sin_theta)

            vertex_edge = Vertex()
            vertex_edge.position = edge
            vertex_edge.normal = normal
            vertex_edge.tex_coord = glm.vec3(0.5+tex_coord_radius*cos_theta, 0.5+tex_coord_radius*sin_theta, 0)

            vertices[i_vertex] = vertex_edge # 1 + j
            i_vertex += 1

            if j > 0:
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 1 + j
                triangle[1] = 0
                triangle[2] = 1 + j - 1
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_sides(self):
        return self.__n_sides
    
    @n_sides.setter
    @Mesh.param_setter
    def n_sides(self, n_sides:int):        
        self.__n_sides = n_sides

    @property
    def start_side(self):
        return self.__start_side

    @start_side.setter
    @Mesh.param_setter
    def start_side(self, start_side:int):
        self.__start_side = start_side

    @property
    def total_sides(self):
        return self.__total_sides

    @total_sides.setter
    @Mesh.param_setter
    def total_sides(self, total_sides:int):
        self.__total_sides = total_sides

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


from .ShaderEffect import ShaderEffect
from glass import sampler2D, ShaderStorageBlock

import time
import os

class ExplosureAdaptor(ShaderEffect):

    class CurrentLuma(ShaderStorageBlock.HostClass):
        def __init__(self):
            ShaderStorageBlock.HostClass.__init__(self)
            self._current_luma = 0

        @property
        def current_luma(self):
            return self._current_luma
        
        @current_luma.setter
        @ShaderStorageBlock.HostClass.not_const
        def current_luma(self, luma:float):
            self._current_luma = luma

    def __init__(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(self, self_folder + "/../glsl/PostProcessEffects/explosure_adaptor.glsl", generate_mipmap=True)

        self.current_luma = ExplosureAdaptor.CurrentLuma()
        self["CurrentLuma"] = self.current_luma

    def apply(self, screen_image:sampler2D)->sampler2D:
        self.program["fps"] = self.camera.screen.smooth_fps
        return ShaderEffect.apply(self, screen_image)
    
    def draw_to_active(self, screen_image: sampler2D) -> None:
        self.program["fps"] = self.camera.screen.smooth_fps
        ShaderEffect.draw_to_active(self, screen_image)

    @property
    def should_update(self) -> bool:
        if not self._enabled:
            return False
        
        if (not self.camera.lens.auto_explosure) or self.camera.lens.local_explosure:
            return False
        
        return (self.screen_update_time == 0 or time.time()-self.screen_update_time <= self.camera.lens.explosure_adapt_time)
    
    @should_update.setter
    def should_update(self, flag:bool):
        pass



from .Renderer import Renderer
from ..PostProcessEffects import GaussBlur, FXAAEffect
from ..Frame import Frame

from glass import ShaderProgram, GLConfig, FBO, sampler2D, sampler2DMS, samplerCube, sampler2DArray, GlassConfig, GLInfo
from glass.utils import cat, modify_time

from OpenGL import GL
import glm
import os

class CommonRenderer(Renderer):

    __dir_light_depth_geo_shader_template = None
    __dir_light_depth_lines_geo_shader_template = None
    __dir_light_depth_points_geo_shader_template = None

    def __init__(self):
        Renderer.__init__(self)

        self._depth_map = None
        self._view_pos_map = None
        self._view_normal_map = None

        self._should_update = False
        self._all_meshes = []
        self._all_lines = []
        self._all_points = []
        self._opaque_meshes = []
        self._opaque_lines = []
        self._opaque_points = []
        self._transparent_meshes = []
        self._transparent_points = []
        self._transparent_lines = []
        self._meshes_cast_shadows = []
        self._lines_cast_shadows = []
        self._points_cast_shadows = []

    @property
    def programs(self):
        if not hasattr(self, "_programs"):
            self._programs = {}
            
        return self._programs
    
    @property
    def fbos(self):
        if not hasattr(self, "_fbos"):
            self._fbos = {}

        return self._fbos
    
    @property
    def dir_light_depth_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename
    
    @property
    def dir_light_depth_lines_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth_lines{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth_lines.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_lines_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_lines_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_lines_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename

    @property
    def dir_light_depth_points_geo_shader_path(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        target_filename = GlassConfig.cache_folder + f"/DirLight_depth_points{self.camera.CSM_levels}.gs"
        template_filename = self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth_points.gs"
        if not os.path.isfile(target_filename) or modify_time(template_filename) > modify_time(target_filename):
            if CommonRenderer.__dir_light_depth_points_geo_shader_template is None:
                CommonRenderer.__dir_light_depth_points_geo_shader_template = cat(template_filename)

            content = CommonRenderer.__dir_light_depth_points_geo_shader_template.replace("{CSM_levels}", str(self.camera.CSM_levels))
            out_file = open(target_filename, "w")
            out_file.write(content)
            out_file.close()

        return target_filename

    @property
    def dir_light_depth_program(self):
        key = f"dir_light_depth{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        program = ShaderProgram()
        program.add_include_path(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/DirLight_depth")
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_geo_shader_path)
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program
    
    @property
    def dir_light_depth_lines_program(self):
        key = f"dir_light_depth_lines{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.add_include_path(self_folder + "/../glsl/Pipelines/DirLight_depth")
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_lines_geo_shader_path)
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program

    @property
    def dir_light_depth_points_program(self):
        key = f"dir_light_depth_points{self.camera.CSM_levels}"
        if key in self.programs:
            return self.programs[key]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.add_include_path(self_folder + "/../glsl/Pipelines/DirLight_depth")
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.vs")
        program.compile(self.dir_light_depth_points_geo_shader_path)
        program.compile(self_folder + "/../glsl/Pipelines/DirLight_depth/DirLight_depth.fs")

        self.programs[key] = program

        return program
    
    @property
    def point_light_depth_program(self):
        if "point_light_depth" in self.programs:
            return self.programs["point_light_depth"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth"] = program

        return program
    
    @property
    def point_light_depth_lines_program(self):
        if "point_light_depth_lines" in self.programs:
            return self.programs["point_light_depth_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth_lines.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth_lines"] = program

        return program

    @property
    def point_light_depth_points_program(self):
        if "point_light_depth_points" in self.programs:
            return self.programs["point_light_depth_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth_points.gs")
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.fs")

        self.programs["point_light_depth_points"] = program

        return program
    
    @property
    def spot_light_depth_program(self):
        if "spot_light_depth" in self.programs:
            return self.programs["spot_light_depth"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth"] = program

        return program
    
    @property
    def spot_light_depth_lines_program(self):
        if "spot_light_depth_lines" in self.programs:
            return self.programs["spot_light_depth_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth_lines.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth_lines"] = program

        return program

    @property
    def spot_light_depth_points_program(self):
        if "spot_light_depth_points" in self.programs:
            return self.programs["spot_light_depth_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/PointLight_depth/PointLight_depth.vs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth_points.gs")
        program.compile(self_folder + "/../glsl/Pipelines/SpotLight_depth/SpotLight_depth.fs")

        self.programs["spot_light_depth_points"] = program

        return program
    
    def classify_meshes(self):
        self._all_meshes.clear()
        self._all_lines.clear()
        self._all_points.clear()
        self._opaque_meshes.clear()
        self._opaque_lines.clear()
        self._opaque_points.clear()
        self._transparent_meshes.clear()
        self._transparent_lines.clear()
        self._transparent_points.clear()
        self._meshes_cast_shadows.clear()
        self._lines_cast_shadows.clear()
        self._points_cast_shadows.clear()

        for mesh_tuple in self.scene.all_meshes.items():
            mesh = mesh_tuple[0]
            if mesh.primitive_type in GLInfo.triangle_types:
                self._all_meshes.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._meshes_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_meshes.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_meshes.append(mesh_tuple)

            elif mesh.primitive_type in GLInfo.line_types:
                self._all_lines.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._lines_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_lines.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_lines.append(mesh_tuple)

            elif mesh.primitive_type == GL.GL_POINTS:
                self._all_points.append(mesh_tuple)
                
                if mesh.material.cast_shadows:
                    self._points_cast_shadows.append(mesh_tuple)

                if mesh.has_transparent:
                    self._transparent_points.append(mesh_tuple)

                if mesh.has_opaque:
                    self._opaque_points.append(mesh_tuple)

    def update_spot_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        if not self._meshes_cast_shadows and \
           not self._lines_cast_shadows and \
           not self._points_cast_shadows:
            return

        for spot_light in self.scene.spot_lights:
            if not spot_light.generate_shadows or \
               not spot_light.need_update_depth_map:
                continue

            if spot_light.depth_fbo is None:
                spot_light.depth_fbo = FBO(1024, 1024)
                spot_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
                spot_light.depth_fbo.depth_attachment.wrap = GL.GL_CLAMP_TO_EDGE
            
            with GLConfig.LocalConfig(depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with spot_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.spot_light_depth_program["spot_light"] = spot_light
                        for mesh, instances in self._meshes_cast_shadows:
                            self.spot_light_depth_program["material"] = mesh.material
                            self.spot_light_depth_program["back_material"] = mesh._back_material
                            self.spot_light_depth_program["explode_distance"] = mesh.explode_distance
                            mesh.draw(self.spot_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.spot_light_depth_lines_program["spot_light"] = spot_light
                        for mesh, instances in self._lines_cast_shadows:
                            self.spot_light_depth_lines_program["material"] = mesh.material
                            self.spot_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.spot_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.spot_light_depth_points_program["spot_light"] = spot_light
                        for mesh, instances in self._points_cast_shadows:
                            self.spot_light_depth_points_program["material"] = mesh.material
                            self.spot_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.spot_light_depth_points_program, instances)

            new_handle = spot_light.depth_fbo.depth_attachment.handle
            if spot_light.depth_map_handle != new_handle:
                spot_light.depth_map_handle = new_handle
                self.scene.spot_lights.dirty = True

            spot_light.need_update_depth_map = False

    def update_point_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        if not self._meshes_cast_shadows and \
           not self._lines_cast_shadows and \
           not self._points_cast_shadows:
            return
        
        for point_light in self.scene.point_lights:
            if not point_light.generate_shadows or \
               not point_light.need_update_depth_map:
                continue

            if point_light.depth_fbo is None:
                point_light.depth_fbo = FBO(1024, 1024)
                point_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
                point_light.depth_fbo.depth_attachment.wrap = GL.GL_CLAMP_TO_EDGE
            
            with GLConfig.LocalConfig(depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with point_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.point_light_depth_program["point_light"] = point_light
                        for mesh, instances in self._meshes_cast_shadows:
                            self.point_light_depth_program["explode_distance"] = mesh.explode_distance
                            self.point_light_depth_program["material"] = mesh.material
                            self.point_light_depth_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.point_light_depth_lines_program["point_light"] = point_light
                        for mesh, instances in self._lines_cast_shadows:
                            self.point_light_depth_lines_program["material"] = mesh.material
                            self.point_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.point_light_depth_points_program["point_light"] = point_light
                        for mesh, instances in self._points_cast_shadows:
                            self.point_light_depth_points_program["material"] = mesh.material
                            self.point_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.point_light_depth_points_program, instances)

            new_handle = point_light.depth_fbo.depth_attachment.handle
            if point_light.depth_map_handle != new_handle:
                point_light.depth_map_handle = new_handle
                self.scene.point_lights.dirty = True

            point_light.need_update_depth_map = False

    def update_dir_lights_depth(self):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return
        
        if not self._meshes_cast_shadows and \
           not self._lines_cast_shadows and \
           not self._points_cast_shadows:
            return
        
        for dir_light in self.scene.dir_lights:
            if not dir_light.generate_shadows:
                continue

            dir_light.max_back_offset = 0

        for mesh, instances in self.scene.all_meshes.items():
            if not mesh.material.cast_shadows:
                continue

            original_center = mesh.center
            original_corner = glm.vec3()
            original_corner.x = original_center.x + (mesh.x_max - mesh.x_min)/2
            original_corner.y = original_center.y + (mesh.y_max - mesh.y_min)/2
            original_corner.z = original_center.z + (mesh.z_max - mesh.z_min)/2
            for instance in instances:
                center = instance.apply(original_center)
                corner = instance.apply(original_corner)
                radius = glm.length(corner - center)

                for dir_light in self.scene.dir_lights:
                    if not dir_light.generate_shadows:
                        continue

                    current_offset = radius + glm.dot(center, -dir_light.direction)
                    if current_offset > dir_light.max_back_offset:
                        dir_light.max_back_offset = current_offset

        for dir_light in self.scene.dir_lights:
            if not dir_light.generate_shadows:
                continue

            if dir_light.depth_fbo is None:
                dir_light.depth_fbo = FBO(1024, 1024, layers=self.camera.CSM_levels)
                dir_light.depth_fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2DArray)

                dir_light.depth_fbo.depth_attachment.wrap_s = GL.GL_CLAMP_TO_BORDER
                dir_light.depth_fbo.depth_attachment.wrap_t = GL.GL_CLAMP_TO_BORDER
                dir_light.depth_fbo.depth_attachment.border_color = glm.vec4(1, 1, 1, 1)

                dir_light.depth_filter = GaussBlur(5)
            
            with GLConfig.LocalConfig(clear_color=glm.vec4(1,1,1,1), depth_test=True, blend=False, cull_face=None, polygon_mode=GL.GL_FILL):
                with dir_light.depth_fbo:
                    if self._meshes_cast_shadows:
                        self.dir_light_depth_program["dir_light"] = dir_light
                        self.dir_light_depth_program["camera"] = self.camera
                        for mesh, instances in self._meshes_cast_shadows:
                            self.dir_light_depth_program["explode_distance"] = mesh.explode_distance
                            self.dir_light_depth_program["material"] = mesh.material
                            self.dir_light_depth_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_program, instances)

                    if self._lines_cast_shadows:
                        self.dir_light_depth_lines_program["dir_light"] = dir_light
                        self.dir_light_depth_lines_program["camera"] = self.camera
                        for mesh, instances in self._lines_cast_shadows:
                            self.dir_light_depth_lines_program["material"] = mesh.material
                            self.dir_light_depth_lines_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_lines_program, instances)

                    if self._points_cast_shadows:
                        self.dir_light_depth_points_program["dir_light"] = dir_light
                        self.dir_light_depth_points_program["camera"] = self.camera
                        for mesh, instances in self._points_cast_shadows:
                            self.dir_light_depth_points_program["material"] = mesh.material
                            self.dir_light_depth_points_program["back_material"] = mesh._back_material
                            mesh.draw(self.dir_light_depth_points_program, instances)

            new_handle = dir_light.depth_fbo.depth_attachment.handle
            if dir_light.depth_map_handle != new_handle:
                dir_light.depth_map_handle = new_handle
                self.scene.dir_lights.dirty = True
                self._should_update = True

    @property
    def forward_program(self):
        if "forward" in self.programs:
            return self.programs["forward"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.gs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_rendering.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward"] = program

        return program
    
    @property
    def forward_lines_program(self):
        if "forward_lines" in self.programs:
            return self.programs["forward_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_lines.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward_lines"] = program

        return program

    @property
    def forward_points_program(self):
        if "forward_points" in self.programs:
            return self.programs["forward_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.vs")
        program.compile(self_folder + "/../glsl/Pipelines/forward_rendering/forward_draw_points.fs")

        program["PointLights"].bind(self.scene.point_lights)
        program["DirLights"].bind(self.scene.dir_lights)
        program["SpotLights"].bind(self.scene.spot_lights)

        self.programs["forward_points"] = program

        return program

    @property
    def OIT_blend_program(self):
        if "OIT_blend" in self.programs:
            return self.programs["OIT_blend"]
        
        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/Pipelines/OIT_blend.fs")
        self.programs["OIT_blend"] = program
        return program
    
    @property
    def OIT_fbo(self):
        fbo = None
        screen_size = GLConfig.screen_size

        samples = self.camera.screen.samples
        if samples > 1:
            if "OIT_ms" in self.fbos:
                fbo = self.fbos["OIT_ms"]
                fbo.resize(screen_size.x, screen_size.y, samples)
            else:
                fbo = FBO(screen_size.x, screen_size.y, samples)
                fbo.attach(0, sampler2DMS, GL.GL_RGBA32F)
                fbo.attach(1, sampler2DMS, GL.GL_RGBA32F)
                fbo.attach(2, sampler2DMS, GL.GL_R32F)
                fbo.attach(3, sampler2DMS, GL.GL_RGB32F)
                fbo.attach(4, sampler2DMS, GL.GL_RGB32F)
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2DMS)
                fbo.auto_clear = False
                self.fbos["OIT_ms"] = fbo
            return fbo
        else:
            if "OIT" in self.fbos:
                fbo = self.fbos["OIT"]
                fbo.resize(screen_size.x, screen_size.y)
            else:
                fbo = FBO(screen_size.x, screen_size.y)
                fbo.attach(0, sampler2D, GL.GL_RGBA32F)
                fbo.attach(1, sampler2D, GL.GL_RGBA32F)
                fbo.attach(2, sampler2D, GL.GL_R32F)
                fbo.attach(3, sampler2D, GL.GL_RGB32F)
                fbo.attach(4, sampler2D, GL.GL_RGB32F)
                fbo.attach(GL.GL_DEPTH_ATTACHMENT, sampler2D)
                fbo.auto_clear = False
                self.fbos["OIT"] = fbo
            return fbo

    def env_OIT_blend_fbo(self, instance):
        current_context = GLConfig.buffered_current_context
        key = ("env_OIT_blend_fbo", current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        fbo = FBO(2048, 1024)
        fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        fbo.color_attachment(0).filter_mipmap = GL.GL_LINEAR
        instance.user_data[key] = fbo
        return fbo

    def env_FXAA_fbo(self, instance):
        current_context = GLConfig.buffered_current_context
        key = ("env_FXAA_fbo", current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        FXAA_fbo = FBO(2048, 1024)
        FXAA_fbo.attach(0, sampler2D)
        FXAA_fbo.color_attachment(0).filter_mipmap = GL.GL_LINEAR
        instance.user_data[key] = FXAA_fbo

        return FXAA_fbo

    @property
    def gen_env_map_program(self):
        if "gen_env_map" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map"] = program

        return self.programs["gen_env_map"]
    
    @property
    def gen_env_map_lines_program(self):
        if "gen_env_map_lines" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_lines.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map_lines"] = program

        return self.programs["gen_env_map_lines"]

    @property
    def gen_env_map_points_program(self):
        if "gen_env_map_points" not in self.programs:
            self_folder = os.path.dirname(os.path.abspath(__file__))
            program = ShaderProgram()
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.vs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.gs")
            program.compile(self_folder + "/../glsl/Pipelines/env_mapping/gen_env_map_points.fs")
            program["DirLights"].bind(self.scene.dir_lights)
            program["PointLights"].bind(self.scene.point_lights)
            program["SpotLights"].bind(self.scene.spot_lights)
            
            self.programs["gen_env_map_points"] = program

        return self.programs["gen_env_map_points"]
    
    def env_map_fbo(self, instance):
        key = ("env_map_fbo", GLConfig.buffered_current_context)
        if key in instance.user_data:
            return instance.user_data[key]
        
        env_map_fbo = FBO(1024, 1024)
        env_map_fbo.attach(0, samplerCube)
        env_map_fbo.attach(1, samplerCube)
        env_map_fbo.attach(2, samplerCube)
        env_map_fbo.attach(GL.GL_DEPTH_ATTACHMENT, samplerCube)
        env_map_fbo.auto_clear = False
        instance.user_data[key] = env_map_fbo
        
        return env_map_fbo

    def env_bake_times(self, instance):
        key = ("env_bake_times", self.scene)
        if key in instance.user_data:
            return instance.user_data[key]
        
        return 0
    
    def increase_bake_times(self, instance):
        key = ("env_bake_times", self.scene)
        if key not in instance.user_data:
            instance.user_data[key] = 0

        instance.user_data[key] += 1

    def prepare_gen_env_map_draw_mesh(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_program["CSM_camera"] = camera
        self.gen_env_map_program["view_center"] = view_center
        self.gen_env_map_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_program["background"] = self.scene.background
        self.gen_env_map_program["fog"] = self.scene.fog

    def gen_env_map_draw_mesh(self, mesh, instances):
        self.gen_env_map_program["explode_distance"] = mesh.explode_distance
        self.gen_env_map_program["material"] = mesh.material
        self.gen_env_map_program["back_material"] = mesh._back_material
        self.gen_env_map_program["is_sphere"] = mesh.is_sphere
        self.gen_env_map_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_program, instances)

    def prepare_gen_env_map_draw_lines(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_lines_program["CSM_camera"] = camera
        self.gen_env_map_lines_program["view_center"] = view_center
        self.gen_env_map_lines_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_lines_program["background"] = self.scene.background
        self.gen_env_map_lines_program["fog"] = self.scene.fog

    def gen_env_map_draw_lines(self, mesh, instances):
        self.gen_env_map_lines_program["material"] = mesh.material
        self.gen_env_map_lines_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_lines_program, instances)

    def prepare_gen_env_map_draw_points(self, view_center:glm.vec3, is_opaque_pass:bool):
        camera = self.camera
        self.gen_env_map_points_program["CSM_camera"] = camera
        self.gen_env_map_points_program["view_center"] = view_center
        self.gen_env_map_points_program["is_opaque_pass"] = is_opaque_pass
        self.gen_env_map_points_program["background"] = self.scene.background
        self.gen_env_map_points_program["fog"] = self.scene.fog

    def gen_env_map_draw_points(self, mesh, instances):
        self.gen_env_map_points_program["material"] = mesh.material
        self.gen_env_map_points_program["mesh_center"] = mesh.center
        mesh.draw(self.gen_env_map_points_program, instances)

    def gen_env_map(self, mesh, instances):
        if "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
            return

        max_bake_times = max(mesh.material.env_max_bake_times, mesh._back_material.env_max_bake_times)
        mesh_center = mesh.center
        for instance in instances:
            if instance.visible == 0:
                continue

            if self.env_bake_times(instance) >= max_bake_times:
                continue
            
            view_center = instance.apply(mesh_center)

            instance.visible = 0
            env_map_fbo = self.env_map_fbo(instance)

            with GLConfig.LocalConfig(depth_test=True, depth_write=True, blend=False):
                with env_map_fbo:
                    GLConfig.clear_buffers()

                    if self._opaque_meshes:
                        self.prepare_gen_env_map_draw_mesh(view_center, True)
                        for other_mesh, other_instances in self._opaque_meshes:
                            self.gen_env_map_draw_mesh(other_mesh, other_instances)

                    if self._opaque_lines:
                        self.prepare_gen_env_map_draw_lines(view_center, True)
                        for other_mesh, other_instances in self._opaque_lines:
                            self.gen_env_map_draw_lines(other_mesh, other_instances)

                    if self._opaque_points:
                        self.prepare_gen_env_map_draw_points(view_center, True)
                        for other_mesh, other_instances in self._opaque_points:
                            self.gen_env_map_draw_points(other_mesh, other_instances)

            opaque_color_map = env_map_fbo.color_attachment(0)
            accum_map = None
            reveal_map = None
            if self._transparent_meshes or self._transparent_points or self._transparent_lines:
                with GLConfig.LocalConfig(
                    depth_test=True, depth_write=False, blend=True,
                    blend_src_rgb=GL.GL_ONE, blend_dest_rgb=GL.GL_ONE,
                    blend_src_alpha=GL.GL_ONE, blend_dest_alpha=GL.GL_ONE):
                    GLConfig.blend_src_rgbi[1] = GL.GL_ONE
                    GLConfig.blend_dest_rgbi[1] = GL.GL_ONE
                    GLConfig.blend_src_rgbi[2] = GL.GL_ONE
                    GLConfig.blend_dest_rgbi[2] = GL.GL_ONE
                    with env_map_fbo:
                        GLConfig.clear_buffer(1, glm.vec4(0))
                        GLConfig.clear_buffer(2, glm.vec4(0))

                        if self._transparent_meshes:
                            self.prepare_gen_env_map_draw_mesh(view_center, False)
                            for other_mesh, other_instances in self._transparent_meshes:
                                self.gen_env_map_draw_mesh(other_mesh, other_instances)

                        if self._transparent_lines:
                            self.prepare_gen_env_map_draw_lines(view_center, False)
                            for other_mesh, other_instances in self._transparent_lines:
                                self.gen_env_map_draw_lines(other_mesh, other_instances)

                        if self._transparent_points:
                            self.prepare_gen_env_map_draw_points(view_center, False)
                            for other_mesh, other_instances in self._transparent_points:
                                self.gen_env_map_draw_points(other_mesh, other_instances)

                    # 取出 OIT 信息
                    accum_map = env_map_fbo.color_attachment(1)
                    reveal_map = env_map_fbo.color_attachment(2)

            env_OIT_blend_fbo = self.env_OIT_blend_fbo(instance)
            with GLConfig.LocalConfig(
                polygon_mode=GL.GL_FILL,
                cull_face=None,
                depth_func=GL.GL_ALWAYS
            ):
                with env_OIT_blend_fbo:
                    self.env_OIT_blend_program["opaque_color_map"] = opaque_color_map
                    self.env_OIT_blend_program["accum_map"] = accum_map
                    self.env_OIT_blend_program["reveal_map"] = reveal_map
                    self.env_OIT_blend_program.draw_triangles(Frame.vertices, Frame.indices)

            env_map = env_OIT_blend_fbo.color_attachment(0)
            instance.user_data["env_bake_state"] = "baked"
            env_FXAA_fbo = self.env_FXAA_fbo(instance)
            with env_FXAA_fbo:
                FXAAEffect.program()["screen_image"] = env_map
                FXAAEffect.program().draw_triangles(Frame.vertices, Frame.indices)
            env_map = env_FXAA_fbo.color_attachment(0)
            instance.user_data["env_bake_state"] = "filtered"
            instance.env_map_handle = env_map.handle
            instance.visible = 1
            self.increase_bake_times(instance)

            if self.env_bake_times(instance) <= max_bake_times:
                self._should_update = True

    def prepare_forward_draw_mesh(self, is_opaque_pass:bool):
        camera = self.camera
        self.forward_program["camera"] = camera
        self.forward_program["is_opaque_pass"] = is_opaque_pass
        self.forward_program["background"] = self.scene.background
        self.forward_program["fog"] = self.scene.fog

    def forward_draw_mesh(self, mesh, instances):
        if mesh.material.need_env_map or mesh._back_material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_program["material"] = mesh.material
        self.forward_program["back_material"] = mesh._back_material
        self.forward_program["explode_distance"] = mesh.explode_distance
        self.forward_program["is_sphere"] = mesh.is_sphere
        self.forward_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_program, instances)

    def prepare_forward_draw_lines(self, is_opaque_pass:bool):
        self.forward_lines_program["camera"] = self.camera
        self.forward_lines_program["is_opaque_pass"] = is_opaque_pass
        self.forward_lines_program["background"] = self.scene.background
        self.forward_lines_program["fog"] = self.scene.fog

    def forward_draw_lines(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_lines_program["material"] = mesh.material
        self.forward_lines_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_lines_program, instances)

    def prepare_forward_draw_points(self, is_opaque_pass:bool):
        self.forward_points_program["camera"] = self.camera
        self.forward_points_program["is_opaque_pass"] = is_opaque_pass
        self.forward_points_program["background"] = self.scene.background
        self.forward_points_program["fog"] = self.scene.fog

    def forward_draw_points(self, mesh, instances):
        if mesh.material.need_env_map:
            self.gen_env_map(mesh, instances)

        self.forward_points_program["material"] = mesh.material
        self.forward_points_program["mesh_center"] = mesh.center
        mesh.draw(self.forward_points_program, instances)

    @property
    def env_OIT_blend_program(self):
        if "env_OIT_blend" in self.programs:
            return self.programs["env_OIT_blend"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(self_folder + "/../glsl/Pipelines/env_mapping/env_OIT_blend.fs")
        self.programs["env_OIT_blend"] = program

        return program

    @property
    def draw_geometry_program(self):
        if "draw_geometry" in self.programs:
            return self.programs["draw_geometry"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.gs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry.fs")
        self.programs["draw_geometry"] = program

        return program
    
    @property
    def draw_geometry_lines_program(self):
        if "draw_geometry_lines" in self.programs:
            return self.programs["draw_geometry_lines"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_lines.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.fs")
        self.programs["draw_geometry_lines"] = program

        return program
    
    @property
    def draw_geometry_points_program(self):
        if "draw_geometry_points" in self.programs:
            return self.programs["draw_geometry_points"]
        
        self_folder = os.path.dirname(os.path.abspath(__file__))
        program = ShaderProgram()
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.vs")
        program.compile(self_folder + "/../glsl/Pipelines/draw_geometry/draw_geometry_points.fs")
        self.programs["draw_geometry_points"] = program

        return program

    def draw_geometry(self, mesh, instances):
        self.draw_geometry_program["material"] = mesh.material
        self.draw_geometry_program["back_material"] = mesh._back_material
        self.draw_geometry_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_program, instances)

    def draw_geometry_lines(self, mesh, instances):
        self.draw_geometry_lines_program["material"] = mesh.material
        self.draw_geometry_lines_program["back_material"] = mesh._back_material
        self.draw_geometry_lines_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_lines_program, instances)

    def draw_geometry_points(self, mesh, instances):
        self.draw_geometry_points_program["material"] = mesh.material
        self.draw_geometry_points_program["back_material"] = mesh._back_material
        self.draw_geometry_points_program["explode_distance"] = mesh.explode_distance
        mesh.draw(self.draw_geometry_points_program, instances)

    def draw_transparent(self):
        if not self._transparent_meshes and \
           not self._transparent_points and \
           not self._transparent_lines:
            return
        
        with GLConfig.LocalConfig(depth_test=True, depth_write=False, blend=True):
            GLConfig.blend_src_rgbi[1] = GL.GL_ONE
            GLConfig.blend_dest_rgbi[1] = GL.GL_ONE
            GLConfig.blend_src_rgbi[2] = GL.GL_ONE
            GLConfig.blend_dest_rgbi[2] = GL.GL_ONE
            with self.OIT_fbo:
                GLConfig.clear_buffer(1, glm.vec4(0))
                GLConfig.clear_buffer(2, glm.vec4(0))

                if self._transparent_meshes:
                    self.prepare_forward_draw_mesh(False)
                    for mesh, instances in self._transparent_meshes:
                        self.forward_draw_mesh(mesh, instances)

                if self._transparent_lines:
                    self.prepare_forward_draw_lines(False)
                    for mesh, instances in self._transparent_lines:
                        self.forward_draw_lines(mesh, instances)

                if self._transparent_points:
                    self.prepare_forward_draw_points(False)
                    for mesh, instances in self._transparent_points:
                        self.forward_draw_points(mesh, instances)

        # 取出 OIT 信息
        resolved = self.OIT_fbo.resolved
        accum_map = resolved.color_attachment(1)
        reveal_map = resolved.color_attachment(2)

        with GLConfig.LocalConfig(
            polygon_mode=GL.GL_FILL,
            cull_face=None,
            depth_func=GL.GL_ALWAYS,
            blend=True,
            blend_func=(GL.GL_ONE_MINUS_SRC_ALPHA, GL.GL_SRC_ALPHA)
        ):
            self.OIT_blend_program["accum_map"] = accum_map
            self.OIT_blend_program["reveal_map"] = reveal_map
            self.OIT_blend_program.draw_triangles(Frame.vertices, Frame.indices)

        if self.camera.screen.post_process_effects.need_pos_info:
            used_fbo = None
            if self.__class__.__name__ == "ForwardRenderer":
                used_fbo = self.OIT_fbo
            elif self.__class__.__name__ == "DeferredRenderer":
                used_fbo = self.gbuffer

            with GLConfig.LocalConfig(depth_test=True, depth_write=True, blend=False):
                with used_fbo:
                    if self._transparent_meshes:
                        self.draw_geometry_program["camera"] = self.camera
                        for mesh, instances in self._transparent_meshes:
                            self.draw_geometry(mesh, instances)

                    if self._transparent_lines:
                        self.draw_geometry_lines_program["camera"] = self.camera
                        for mesh, instances in self._transparent_lines:
                            self.draw_geometry_lines(mesh, instances)

                    if self._transparent_points:
                        self.draw_geometry_points_program["camera"] = self.camera
                        for mesh, instances in self._transparent_points:
                            self.draw_geometry_points(mesh, instances)

            resolved = used_fbo.resolved
            self._depth_map = resolved.depth_attachment
            self._view_pos_map = resolved.color_attachment(3)
            self._view_normal_map = resolved.color_attachment(4)



from OpenGL import GL
import OpenGL.GL.ARB.gpu_shader_int64 as gsi64
import ctypes
import functools

from .helper import nitems, sizeof
from .GLObject import GLObject
from .GLInfo import GLInfo
from .GLConfig import GLConfig

def contex_check(func):
	@functools.wraps(func)
	def wraps(*args, **kwargs):
		self = args[0]
		cls = self.__class__
		if self._vao.context != 0 and self._vao.context != GLConfig.buffered_current_context:
			if self._vao.context not in cls._cmd_buffer:
				cls._cmd_buffer[self._vao.context] = []
			cls._cmd_buffer[self._vao.context].append((func, args, kwargs))
			return
		
		return func(*args, **kwargs)
	
	return wraps

class VAP:

	_cmd_buffer = {}

	def __init__(self, vao, location):
		self._vao = vao
		self._location = location
		self._vbo = None
		self._element_type = None
		self._stride = 0
		self._offset = 0
		self._divisor = 0
		self._enabled = False

	@contex_check
	def interp(self, vbo, element_type, stride = 0, offset = 0):
		self._vao.bind()
		vbo.bind()
		if stride == 0:
			stride = sizeof(element_type)

		gl_type = GLInfo.dtype_inverse_map[element_type]
		if gl_type in [GL.GL_DOUBLE, gsi64.GL_UNSIGNED_INT64_ARB]:
			GL.glVertexAttribLPointer(self._location, nitems(element_type), gl_type, stride, ctypes.c_void_p(offset))
		elif gl_type in GLInfo.int_types:
			GL.glVertexAttribIPointer(self._location, nitems(element_type), gl_type, stride, ctypes.c_void_p(offset))
		else:
			GL.glVertexAttribPointer(self._location, nitems(element_type), gl_type, GL.GL_FALSE, stride, ctypes.c_void_p(offset))
		self.enabled = True

		self._vbo = vbo
		self._element_type = element_type
		self._stride = stride
		self._offset = offset

	@property
	def element_type(self):
		return self._element_type

	@property
	def stride(self):
		return self._stride

	@property
	def offset(self):
		return self._offset

	@property
	def divisor(self):
		return self._divisor
	
	@property
	def enabled(self):
		return self._enabled
	
	@enabled.setter
	@contex_check
	def enabled(self, flag:bool):
		self._vao.bind()
		if flag:
			GL.glEnableVertexAttribArray(self._location)
		else:
			GL.glDisableVertexAttribArray(self._location)
		self._enabled = flag

	@divisor.setter
	@contex_check
	def divisor(self, value:int):
		if value < 0:
			raise ValueError("divisor should be integer that not less than 0")

		if self._divisor == value:
			return
		
		self._vao.bind()
		GL.glVertexAttribDivisor(self._location, value)
		self._divisor = value
		self.enabled = True

class VAO(GLObject):

	_basic_info = \
	{
		"gen_func": GL.glGenVertexArrays,
		"bind_func": GL.glBindVertexArray,
		"del_func": GL.glDeleteVertexArrays,
		"target_type": None,
		"binding_type": GL.GL_VERTEX_ARRAY_BINDING,
		"need_number": True,
	}

	def __init__(self):
		GLObject.__init__(self, context_shared=False)
		self._ebo = None
		self._VAP_map = {}
		self._context:int = 0

	def __getitem__(self, location):
		if location not in self._VAP_map:
			self._VAP_map[location] = VAP(self, location)
			
		return self._VAP_map[location]
	
	def __contains__(self, location):
		return (location in self._VAP_map)
	
	@staticmethod
	def execute_cmd_buffer():
		current_context = GLConfig.buffered_current_context
		if current_context not in VAP._cmd_buffer:
			return
		
		for cmd in VAP._cmd_buffer[current_context]:
			cmd[0](*cmd[1], **cmd[2])

	def setEBO(self, ebo):
		if self._ebo is ebo:
			return
			
		self.bind()
		ebo.bind()
		self._ebo = ebo

	@property
	def ebo(self):
		return self._ebo
	
	@property
	def context(self):
		return self._context


from .WeakRef import WeakRef

class WeakDict:

    class iterator:
        def __init__(self, weak_ref,  _iterator):
            self._iterator = _iterator
            self._weak_ref = weak_ref

        def __next__(self):
            value = self._iterator.__next__()
            if self._weak_ref:
                value = value.value

            return value
        
        def __iter__(self):
            return self
        
    class items_iterator:
        def __init__(self, weak_ref_keys, weak_ref_values, _iterator):
            self._iterator = _iterator
            self._weak_ref_keys = weak_ref_keys
            self._weak_ref_values = weak_ref_values

        def __next__(self):
            key, value = self._ref_iterator.__next__()
            if self._weak_ref_keys:
                key = key.value
            if self._weak_ref_values:
                value = value.value
            return key, value
        
        def __iter__(self):
            return self

    def __init__(self, weak_ref_keys=True, weak_ref_values=True):
        self._dict = {}
        self._weak_ref_keys = weak_ref_keys
        self._weak_ref_values = weak_ref_values

    def __iter__(self):
        result = WeakDict.iterator(self._weak_ref_keys, self._dict.__iter__())
        return result
    
    def __len__(self):
        return len(self._dict)
    
    def __bool__(self):
        return bool(self._dict)
    
    def __contains__(self, key):
        if self._weak_ref_keys:
            key = WeakRef(key)
            
        return key in self._dict
    
    def __getitem__(self, key):
        if self._weak_ref_keys:
            key = WeakRef(key)

        value = self._dict[key]
        if self._weak_ref_values:
            value = value.value
            
        return value

    def __setitem__(self, key, value):
        if self._weak_ref_keys:
            key = WeakRef(key)

        if self._weak_ref_values:
            value = WeakRef(value)

        self._dict[key] = value

    def __delitem__(self, key):
        if self._weak_ref_keys:
            key = WeakRef(key)

        del self._dict[key]

    def __repr__(self):
        return self._dict.__repr__()

    def pop(self, key):
        if self._weak_ref_keys:
            key = WeakRef(key)

        value = self._dict.pop(key)
        if self._weak_ref_values:
            value = value.value
            
        return value

    def clear(self):
        self._dict.clear()

    def update(self, _dict):
        for key, value in _dict.items():
            self[key] = value
    
    def keys(self):
        result = WeakDict.iterator(self._weak_ref_keys, self._dict.keys())
        return result
    
    def values(self):
        result = WeakDict.iterator(self._weak_ref_values, self._dict.values())
        return result
    
    def items(self):
        result = WeakDict.items_iterator(self._weak_ref_keys, self._weak_ref_values, self._dict.items())
        return result
    



from OpenGL import GL
import numpy as np

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype
from .sampler2D import sampler2D

class image2D(sampler2D):

    _default_internal_format = GL.GL_RGBA32F
    _default_filter_min = GL.GL_NEAREST
    _default_filter_mag = GL.GL_NEAREST
    _default_filter_mipmap = None

    @checktype
    def __init__(self, image:(str,np.ndarray)=None, width:int=None, height:int=None, internal_format:GLInfo.image_internal_formats=None):
        sampler2D.__init__(self, image, width, height, internal_format)

    @property
    def internal_format(self):
        return self._internal_format

    @internal_format.setter
    @FBOAttachment.param_setter
    def internal_format(self, internal_format:GLInfo.image_internal_formats):
        self._set_internal_format(internal_format)

    @property
    def filter_min(self):
        return GL.GL_NEAREST

    @filter_min.setter
    def filter_min(self, filter_type:GLInfo.filter_types):
        raise RuntimeError("cannot set filter for image2D")

    @property
    def filter_mag(self):
        return GL.GL_NEAREST

    @filter_mag.setter
    def filter_mag(self, filter_type:GLInfo.filter_types):
        raise RuntimeError("cannot set filter for image2D")

    @property
    def filter_mipmap(self):
        return GL.GL_NONE

    @filter_mipmap.setter
    def filter_mipmap(self, filter_type:GLInfo.filter_types):
        raise RuntimeError("cannot set filter for image2D")
        
    @property
    def filter(self):
        return GL.GL_NEAREST, GL.GL_NEAREST, None

    @filter.setter
    def filter(self, filter_type):
        raise RuntimeError("cannot set filter for image2D")


import OpenGL
from OpenGL import GL
import glm
import numpy as np
import platform
import inspect
import moderngl

from .GLInfo import GLInfo
from .helper import glGetEnum, glGetEnumi

op_system = platform.system()
getCurrentContext = None
if op_system == 'Windows':
    from OpenGL import WGL
    getCurrentContext = WGL.wglGetCurrentContext
elif op_system == 'Linux':
    from OpenGL import GLX
    getCurrentContext = GLX.glXGetCurrentContext

class StencilFunc:

    def __init__(self, func:[*GLInfo.stencil_funcs, *GLInfo.stencil_func_strs]=None, ref:int=None, mask:int=None):
        if func is None:
            func = glGetEnum(GL.GL_STENCIL_FUNC)

        if ref is None:
            ref = glGetEnum(GL.GL_STENCIL_REF)

        if mask is None:
            mask = glGetEnum(GL.GL_STENCIL_VALUE_MASK)

        if isinstance(func, str):
            func = GLInfo.stencil_func_map[func]

        self.__func = func
        self.__ref = ref
        self.__mask = mask

    def apply(self):
        GL.glStencilFunc(self.__func, self.__ref, self.__mask)

    def update(self):
        self.__ref = glGetEnum(GL.GL_STENCIL_REF)
        self.__mask = glGetEnum(GL.GL_STENCIL_VALUE_MASK)
        self.__func = glGetEnum(GL.GL_STENCIL_FUNC)

    @property
    def func(self):
        return self.__func
    
    @property
    def ref(self):
        return self.__ref
    
    @property
    def mask(self):
        return self.__mask

    def __and__(self, mask:int):
        self.__mask = mask

    def __lt__(self, ref):
        GL.glStencilFunc(GL.GL_LESS, (~ref) | self.__mask, self.__mask)
        self.__func = "<"
        self.__ref = ref

    def __le__(self, ref):
        GL.glStencilFunc(GL.GL_LEQUAL, (~ref) | self.__mask, self.__mask)
        self.__func = "<="
        self.__ref = ref
    
    def __eq__(self, ref):
        if ref.__class__ == int:
            GL.glStencilFunc(GL.GL_EQUAL, (~ref) | self.__mask, self.__mask)
            self.__func = "=="
            self.__ref = ref
        elif isinstance(ref, self.__class__):
            return (self.__func == ref.__func and \
                    self.__ref == ref.__ref and \
                    self.__mask == ref.__mask)
        elif isinstance(ref, tuple):
            return (self.__func == ref[0] and \
                    self.__ref == ref[1] and \
                    self.__mask == ref[2])
        else:
            return False
    
    def __gt__(self, ref):
        GL.glStencilFunc(GL.GL_GREATER, (~ref) | self.__mask, self.__mask)
        self.__func = ">"
        self.__ref = ref
    
    def __ge__(self, ref):
        GL.glStencilFunc(GL.GL_GEQUAL, (~ref) | self.__mask, self.__mask)
        self.__func = ">="
        self.__ref = ref
    
    def __ne__(self, ref):
        if ref.__class__ == int:
            GL.glStencilFunc(GL.GL_NOTEQUAL, (~ref) | self.__mask, self.__mask)
            self.__func = "!="
            self.__ref = ref
        elif isinstance(ref, self.__class__):
            return (self.__func != ref.__func or \
                    self.__ref != ref.__ref or \
                    self.__mask != ref.__mask)
        else:
            return True

class _MetaGLConfig(type):
    
    class BlendSrcRGBi:
        def __getitem__(self, index:int):
            return glGetEnumi(GL.GL_BLEND_SRC_RGB, index)
        
        def __setitem__(self, index:int, value:GLInfo.blend_funcs):
            GL.glBlendFunci(index, value, glGetEnumi(GL.GL_BLEND_DST_RGB, index))
    BlendSrcRGBi = BlendSrcRGBi()

    class BlendDstRGBi:
        def __getitem__(self, index:int):
            return glGetEnumi(GL.GL_BLEND_DST_RGB, index)
        
        def __setitem__(self, index:int, value:GLInfo.blend_funcs):
            GL.glBlendFunci(index, glGetEnumi(GL.GL_BLEND_SRC_RGB, index), value)
    BlendDstRGBi = BlendDstRGBi()

    class BlendSrcAlphai:
        def __getitem__(self, index:int):
            return glGetEnumi(GL.GL_BLEND_SRC_ALPHA, index)
        
        def __setitem__(self, index:int, value:GLInfo.blend_funcs):
            GL.glBlendFunci(index, value, glGetEnumi(GL.GL_BLEND_DST_ALPHA, index))
    BlendSrcAlphai = BlendSrcAlphai()

    class BlendDstAlphai:
        def __getitem__(self, index:int):
            return glGetEnumi(GL.GL_BLEND_DST_ALPHA, index)
        
        def __setitem__(self, index:int, value:GLInfo.blend_funcs):
            GL.glBlendFunci(index, glGetEnumi(GL.GL_BLEND_SRC_ALPHA, index), value)
    BlendDstAlphai = BlendDstAlphai()

    __debug = True

    __max_texture_units = None
    __max_image_units = None
    __max_compute_work_group_count = None
    __max_compute_work_group_size = None
    __max_compute_work_group_invocations = None
    __max_uniform_buffer_bindings = None
    __max_shader_storage_buffer_bindings = None
    __max_color_attachments = None
    __max_draw_buffers = None
    __screen_size = {}
    __buffered_current_context = None
    __buffered_viewport = {}
    __gl_version = None
    __gl_major_version = None
    __gl_minor_version = None
    __gl_renderer = None
    __available_texture_units = None
    __available_image_units = None
    __available_extensions = None
    __depth_bits = None
    __standalone_context = None

    @property
    def debug(cls):
        return _MetaGLConfig.__debug
    
    @debug.setter
    def debug(cls, flag:bool):
        if _MetaGLConfig.__debug == flag:
            return
        
        _MetaGLConfig.__debug = flag
        OpenGL.ERROR_CHECKING = flag

    @property
    def stencil_test(cls):
        return bool(GL.glIsEnabled(GL.GL_STENCIL_TEST))
    
    @stencil_test.setter
    def stencil_test(cls, flag:bool):
        if flag:
            GL.glEnable(GL.GL_STENCIL_TEST)
        else:
            GL.glDisable(GL.GL_STENCIL_TEST)

    @property
    def stencil_mask(cls):
        return glGetEnum(GL.GL_STENCIL_VALUE_MASK)
    
    @stencil_mask.setter
    def stencil_mask(cls, mask:int):
        GL.glStencilMask(mask)

    @property
    def stencil_func(cls):
        return StencilFunc()
    
    @stencil_func.setter
    def stencil_func(cls, func):
        if isinstance(func, StencilFunc):
            func.apply()
            return

        if func in [True, "always", GL.GL_ALWAYS]:
            GL.glStencilFunc(GL.GL_ALWAYS, 1, 0xFF)
        elif func in [False, None, "never", GL.GL_NEVER]:
            GL.glStencilFunc(GL.GL_NEVER, 1, 0xFF)
        else:
            f = func[0]
            if isinstance(f, str):
                f = GLInfo.stencil_func_map[f]
            GL.glStencilFunc(f, func[1], func[2])

    @property
    def stencil_fail(cls):
        return glGetEnum(GL.GL_STENCIL_FAIL)
    
    @stencil_fail.setter
    def stencil_fail(cls, operation:GLInfo.operations):
        GL.glStencilOp(operation, cls.stencil_pass_depth_fail, cls.stencil_pass_depth_pass)

    @property
    def stencil_pass_depth_fail(cls):
        return glGetEnum(GL.GL_STENCIL_PASS_DEPTH_FAIL)
    
    @stencil_pass_depth_fail.setter
    def stencil_pass_depth_fail(cls, operation:GLInfo.operations):
        GL.glStencilOp(cls.stencil_fail, operation, cls.stencil_pass_depth_pass)

    @property
    def stencil_pass_depth_pass(cls):
        return glGetEnum(GL.GL_STENCIL_PASS_DEPTH_PASS)
    
    @stencil_pass_depth_pass.setter
    def stencil_pass_depth_pass(cls, operation:GLInfo.operations):
        GL.glStencilOp(cls.stencil_fail, cls.stencil_pass_depth_fail, operation)

    @property
    def stencil_op(cls):
        return (cls.stencil_fail, cls.stencil_pass_depth_fail, cls.stencil_pass_depth_pass)
    
    @stencil_op.setter
    def stencil_op(cls, stencil_op):
        GL.glStencilOp(*stencil_op)

    @property
    def depth_test(cls):
        return bool(GL.glIsEnabled(GL.GL_DEPTH_TEST))
    
    @depth_test.setter
    def depth_test(cls, flag:bool):
        if flag:
            GL.glEnable(GL.GL_DEPTH_TEST)
        else:
            GL.glDisable(GL.GL_DEPTH_TEST)

    @property
    def depth_mask(cls):
        result = GL.glGetBooleanv(GL.GL_DEPTH_WRITEMASK)
        return bool(result)
    
    @depth_mask.setter
    def depth_mask(cls, flag:bool):
        if flag:
            GL.glDepthMask(GL.GL_TRUE)
        else:
            GL.glDepthMask(GL.GL_FALSE)

    @property
    def depth_write(cls):
        result = GL.glGetBooleanv(GL.GL_DEPTH_WRITEMASK)
        return bool(result)
    
    @depth_write.setter
    def depth_write(cls, flag:bool):
        if flag:
            GL.glDepthMask(GL.GL_TRUE)
        else:
            GL.glDepthMask(GL.GL_FALSE)

    @property
    def depth_func(cls):
        depth_func = glGetEnum(GL.GL_DEPTH_FUNC)
        return GLInfo.depth_func_map[depth_func]
        
    @depth_func.setter
    def depth_func(cls, depth_func):
        if depth_func not in GLInfo.depth_funcs:
            depth_func = GLInfo.depth_func_map[depth_func]

        GL.glDepthFunc(depth_func)

    @property
    def blend(cls):
        return bool(GL.glIsEnabled(GL.GL_BLEND))
    
    @blend.setter
    def blend(cls, flag:bool):
        if flag:
            GL.glEnable(GL.GL_BLEND)
        else:
            GL.glDisable(GL.GL_BLEND)

    @property
    def blend_equation(cls):
        blend_equation = glGetEnum(GL.GL_BLEND_EQUATION)
        return GLInfo.blend_equation_map[blend_equation]
    
    @blend_equation.setter
    def blend_equation(cls, blend_equation:list(GLInfo.blend_equation_map.keys())):
        if isinstance(blend_equation, str):
            blend_equation = GLInfo.blend_equation_map[blend_equation]

        GL.glBlendEquation(blend_equation)

    @property
    def blend_func(cls):
        return cls.blend_src_rgb, cls.blend_dest_rgb
    
    @blend_func.setter
    def blend_func(cls, args:tuple):
        GL.glBlendFunc(*args)

    @property
    def blend_src_rgbi(cls):
        return _MetaGLConfig.BlendSrcRGBi
    
    @property
    def blend_dest_rgbi(cls):
        return _MetaGLConfig.BlendDstRGBi
    
    @property
    def blend_src_alphai(cls):
        return _MetaGLConfig.BlendSrcAlphai
    
    @property
    def blend_dest_alphai(cls):
        return _MetaGLConfig.BlendDstAlphai

    @property
    def blend_src_rgb(cls):
        return glGetEnum(GL.GL_BLEND_SRC_RGB)
    
    @blend_src_rgb.setter
    def blend_src_rgb(cls, value:GLInfo.blend_funcs):
        if value == cls.blend_src_rgb:
            return
        
        GL.glBlendFunc(value, cls.blend_dest_rgb)
    
    @property
    def blend_dest_rgb(cls):
        return glGetEnum(GL.GL_BLEND_DST_RGB)

    @blend_dest_rgb.setter
    def blend_dest_rgb(cls, value:GLInfo.blend_funcs):
        if value == cls.blend_dest_rgb:
            return
        
        GL.glBlendFunc(cls.blend_src_rgb, value)
    
    @property
    def blend_src_alpha(cls):
        return glGetEnum(GL.GL_BLEND_SRC_ALPHA)
    
    @blend_src_alpha.setter
    def blend_src_alpha(cls, value:GLInfo.blend_funcs):
        if value == cls.blend_src_alpha:
            return
        
        GL.glBlendFuncSeparate(cls.blend_src_rgb, cls.blend_dest_rgb, value, cls.blend_dest_alpha)
    
    @property
    def blend_dest_alpha(cls):
        return glGetEnum(GL.GL_BLEND_DST_ALPHA)

    @blend_dest_alpha.setter
    def blend_dest_alpha(cls, value:GLInfo.blend_funcs):
        if value == cls.blend_dest_alpha:
            return
        
        GL.glBlendFuncSeparate(cls.blend_src_rgb, cls.blend_dest_rgb, cls.blend_src_alpha, value)

    @property
    def cull_face(cls):
        if not GL.glIsEnabled(GL.GL_CULL_FACE):
            return None
        
        return glGetEnum(GL.GL_CULL_FACE_MODE)

    @cull_face.setter
    def cull_face(cls, face:GLInfo.cull_face_types):
        if face is None:
            GL.glDisable(GL.GL_CULL_FACE)
        else:
            GL.glEnable(GL.GL_CULL_FACE)
            GL.glCullFace(face)

    @property
    def clear_color(cls):
        color_array = np.array([0.0]*4, dtype=np.float32)
        GL.glGetFloatv(GL.GL_COLOR_CLEAR_VALUE, color_array)
        return glm.vec4(color_array[0], color_array[1], color_array[2], color_array[3])
    
    @clear_color.setter
    def clear_color(cls, color:(tuple,glm.vec3,glm.vec4,float)):
        if isinstance(color, float):
            GL.glClearColor(color, color, color, color)
        elif len(color) == 3:
            GL.glClearColor(color[0], color[1], color[2], 1)
        elif len(color) == 4:
            GL.glClearColor(color[0], color[1], color[2], color[3])
        else:
            raise ValueError(color)

    @property
    def polygon_mode(cls):
        return glGetEnum(GL.GL_POLYGON_MODE)
    
    @polygon_mode.setter
    def polygon_mode(cls, mode:GLInfo.polygon_modes):
        GL.glPolygonMode(GL.GL_FRONT_AND_BACK, mode)

    @property
    def line_width(cls):
        return GL.glGetFloatv(GL.GL_LINE_WIDTH)
    
    @line_width.setter
    def line_width(cls, line_width:float):
        GL.glLineWidth(line_width)

    @property
    def point_size(cls):
        return GL.glGetFloatv(GL.GL_POINT_SIZE)
    
    @point_size.setter
    def point_size(cls, point_size:float):
        GL.glPointSize(point_size)

    @property
    def viewport(cls):
        current_viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
        return (int(current_viewport[0]), int(current_viewport[1]), int(current_viewport[2]), int(current_viewport[3]))

    @viewport.setter
    def viewport(cls, viewport:tuple):
        GL.glViewport(*viewport)

    @property
    def buffered_viewport(cls):
        current_context = cls.buffered_current_context
        if current_context not in _MetaGLConfig.__buffered_viewport:
            return cls.viewport
        
        return _MetaGLConfig.__buffered_viewport[current_context]
    
    @buffered_viewport.setter
    def buffered_viewport(cls, viewport):
        current_context = cls.buffered_current_context
        _MetaGLConfig.__buffered_viewport[current_context] = viewport

    @property
    def screen_size(cls)->glm.ivec2:
        current_context = cls.buffered_current_context
        if current_context not in _MetaGLConfig.__screen_size:
            _MetaGLConfig.__screen_size[current_context] = glm.ivec2()

        viewport = cls.viewport
        _MetaGLConfig.__screen_size[current_context].x = viewport[2]
        _MetaGLConfig.__screen_size[current_context].y = viewport[3]
        return _MetaGLConfig.__screen_size[current_context]

    @property
    def active_texture_unit(cls):
        return int(GL.glGetIntegerv(GL.GL_ACTIVE_TEXTURE) - GL.GL_TEXTURE0)
    
    @active_texture_unit.setter
    def active_texture_unit(cls, unit):
        if unit >= cls.max_texture_units:
            raise ValueError(f"try to active texture unit {unit}, which is over hardware available texture units range 0~{cls.max_texture_units-1}")
        
        GL.glActiveTexture(GL.GL_TEXTURE0 + unit)

    # Read only:
    @property
    def depth_bits(self):
        if _MetaGLConfig.__depth_bits is None:
            _MetaGLConfig.__depth_bits = int(GL.glGetIntegerv(GL.GL_DEPTH_BITS))

        return _MetaGLConfig.__depth_bits

    @property
    def max_color_attachments(cls):
        if _MetaGLConfig.__max_color_attachments is None:
            try:
                _MetaGLConfig.__max_color_attachments = GL.glGetIntegerv(GL.GL_MAX_COLOR_ATTACHMENTS)
            except:
                return 8
            
        return _MetaGLConfig.__max_color_attachments
    
    @property
    def max_draw_buffers(cls):
        if _MetaGLConfig.__max_draw_buffers is None:
            try:
                _MetaGLConfig.__max_draw_buffers = GL.glGetIntegerv(GL.GL_MAX_DRAW_BUFFERS)
            except:
                return 8
            
        return _MetaGLConfig.__max_draw_buffers
    
    @property
    def max_texture_units(cls):
        if _MetaGLConfig.__max_texture_units is None:
            _MetaGLConfig.__max_texture_units = GL.glGetIntegerv(GL.GL_MAX_TEXTURE_IMAGE_UNITS)
        return _MetaGLConfig.__max_texture_units
    
    @property
    def max_image_units(cls):
        if _MetaGLConfig.__max_image_units is None:
            _MetaGLConfig.__max_image_units = GL.glGetIntegerv(GL.GL_MAX_IMAGE_UNITS)
        return _MetaGLConfig.__max_image_units

    @property
    def max_compute_work_group_count(cls):
        if _MetaGLConfig.__max_compute_work_group_count is None:
            max_x = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 0)[0]
            max_y = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 1)[0]
            max_z = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_COUNT, 2)[0]
            _MetaGLConfig.__max_compute_work_group_count = (max_x, max_y, max_z)

        return _MetaGLConfig.__max_compute_work_group_count

    @property
    def max_compute_work_group_size(cls):
        if _MetaGLConfig.__max_compute_work_group_size is None:
            max_x = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 0)[0]
            max_y = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 1)[0]
            max_z = GL.glGetIntegeri_v(GL.GL_MAX_COMPUTE_WORK_GROUP_SIZE, 2)[0]
            _MetaGLConfig.__max_compute_work_group_size = (max_x, max_y, max_z)
        return _MetaGLConfig.__max_compute_work_group_size

    @property
    def max_compute_work_group_invocations(cls):
        if _MetaGLConfig.__max_compute_work_group_invocations is None:
            _MetaGLConfig.__max_compute_work_group_invocations = GL.glGetIntegerv(GL.GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS)
        return _MetaGLConfig.__max_compute_work_group_invocations

    @property
    def max_uniform_buffer_bindings(cls):
        if _MetaGLConfig.__max_uniform_buffer_bindings is None:
            _MetaGLConfig.__max_uniform_buffer_bindings = int(GL.glGetIntegerv(GL.GL_MAX_UNIFORM_BUFFER_BINDINGS))
        return _MetaGLConfig.__max_uniform_buffer_bindings

    @property
    def max_shader_storage_buffer_bindings(cls):
        if _MetaGLConfig.__max_shader_storage_buffer_bindings is None:
            _MetaGLConfig.__max_shader_storage_buffer_bindings = int(GL.glGetIntegerv(GL.GL_MAX_SHADER_STORAGE_BUFFER_BINDINGS))
        return _MetaGLConfig.__max_shader_storage_buffer_bindings

    @property
    def current_context(cls):
        return getCurrentContext()
        
    @property
    def buffered_current_context(cls):
        if _MetaGLConfig.__buffered_current_context is None:
            return cls.current_context
        else:
            return _MetaGLConfig.__buffered_current_context
        
    @buffered_current_context.setter
    def buffered_current_context(cls, context_id:int):
        _MetaGLConfig.__buffered_current_context = context_id
    
    @property
    def standalone_context(cls)->moderngl.Context:
        if cls.__standalone_context is None:
            cls.__standalone_context = moderngl.create_standalone_context(require=430)

        return cls.__standalone_context
    
    @property
    def version(cls)->str:
        if _MetaGLConfig.__gl_version is None:
            if cls.buffered_current_context is None:
                _MetaGLConfig.__gl_version = cls.standalone_context.info["GL_VERSION"]
            else:
                _MetaGLConfig.__gl_version = GL.glGetString(GL.GL_VERSION).decode("utf-8")
        
        return _MetaGLConfig.__gl_version
    
    @property
    def major_version(cls)->str:
        if _MetaGLConfig.__gl_major_version is None:
            if cls.buffered_current_context is None:
                _MetaGLConfig.__gl_major_version = cls.standalone_context.info["GL_MAJOR_VERSION"]
            else:
                _MetaGLConfig.__gl_major_version = GL.glGetInteger(GL.GL_MAJOR_VERSION)
        
        return _MetaGLConfig.__gl_major_version
    
    @property
    def minor_version(cls)->str:
        if _MetaGLConfig.__gl_minor_version is None:
            if cls.buffered_current_context is None:
                _MetaGLConfig.__gl_minor_version = cls.standalone_context.info["GL_MINOR_VERSION"]
            else:
                _MetaGLConfig.__gl_minor_version = GL.glGetInteger(GL.GL_MINOR_VERSION)
        
        return _MetaGLConfig.__gl_minor_version

    @property
    def renderer(cls)->str:
        if _MetaGLConfig.__gl_renderer is None:
            if cls.buffered_current_context is None:
                _MetaGLConfig.__gl_renderer = cls.standalone_context.info["GL_RENDERER"]
            else:
                _MetaGLConfig.__gl_renderer = GL.glGetString(GL.GL_RENDERER).decode("utf-8")
        
        return _MetaGLConfig.__gl_renderer

    @property
    def available_texture_units(cls):
        if _MetaGLConfig.__available_texture_units is None:
            _MetaGLConfig.__available_texture_units = set(range(cls.max_texture_units))

        return _MetaGLConfig.__available_texture_units
    
    @property
    def available_image_units(cls):
        if _MetaGLConfig.__available_image_units is None:
            _MetaGLConfig.__available_image_units = set(range(cls.max_image_units))

        return _MetaGLConfig.__available_image_units
    
    @property
    def available_extensions(cls):
        if _MetaGLConfig.__available_extensions is None:
            extensions = GL.glGetString(GL.GL_EXTENSIONS).decode("utf-8")
            _MetaGLConfig.__available_extensions = extensions.split()

        return _MetaGLConfig.__available_extensions
    
def setter_methods(cls):
    setters = set()
    for name, member in inspect.getmembers(cls):
        if isinstance(member, property) and member.fset is not None:
            setters.add(name)
    return setters

class GLConfig(metaclass=_MetaGLConfig):

    class LocalConfig(dict):
        __all_attrs = {*dir({}), "__getattr__", "__enter__", "__exit__", "_old_values"}
        __setters = setter_methods(_MetaGLConfig)

        def __init__(self, **kwargs):
            dict.__init__(self, **kwargs)
            self._old_values = {}

        def __getattr__(self, name):
            if name in GLConfig.LocalConfig.__all_attrs:
                return dict.__getattr__(self, name)
            
            return dict.__getitem__(self, name)
        
        def __setitem__(self, name, value):
            if name not in GLConfig.LocalConfig.__setters:
                raise AttributeError(f"'{name}' is not an attribute of GLConfig")
            
            dict.__setitem__(self, name, value)

        def __setattr__(self, name, value):
            if name in GLConfig.LocalConfig.__all_attrs:
                return dict.__setattr__(self, name, value)
            
            self.__setitem__(name, value)
        
        def __delattr__(self, name):
            if name in GLConfig.LocalConfig.__all_attrs:
                return dict.__delattr__(self, name)
            
            if name in self:
                dict.__delitem__(self, name)
        
        def __enter__(self):
            self._old_values.clear()
            for name in self:
                old_value = getattr(GLConfig, name)
                new_value = self[name]
                if old_value != new_value:
                    self._old_values[name] = old_value
                    setattr(GLConfig, name, new_value)

        def __exit__(self, *exc_details):
            for name, value in self._old_values.items():
                setattr(GLConfig, name, value)

    @staticmethod
    def clear_buffers(bits=None):
        if bits is None:
            bits = (GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT)
        
        GL.glClear(bits)

    @staticmethod
    def clear_buffer(target, color:(tuple,float,glm.vec3,glm.vec4)):
        np_color = np.array([0]*4)
        if isinstance(color, float):
            np_color[0] = color
            np_color[1] = color
            np_color[2] = color
            np_color[3] = color
        elif len(color) == 3:
            np_color[0] = color[0]
            np_color[1] = color[1]
            np_color[2] = color[2]
            np_color[3] = 1
        elif len(color) == 4:
            np_color[0] = color[0]
            np_color[1] = color[1]
            np_color[2] = color[2]
            np_color[3] = color[3]
        else:
            raise ValueError(color)
        
        if 0 <= target < GLConfig.max_color_attachments:
            GL.glClearBufferfv(GL.GL_COLOR, target, np_color)
        elif GL.GL_COLOR_ATTACHMENT0 <= target < GLConfig.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
            GL.glClearBufferfv(GL.GL_COLOR, target - GL.GL_COLOR_ATTACHMENT0, np_color)
        elif target in [GL.GL_DEPTH, GL.GL_DEPTH_BUFFER_BIT, GL.GL_DEPTH_ATTACHMENT]:
            GL.glClearBufferfv(GL.GL_DEPTH, 0, np_color)
        elif target in [GL.GL_STENCIL, GL.GL_STENCIL_BUFFER_BIT, GL.GL_STENCIL_ATTACHMENT]:
            GL.glClearBufferfv(GL.GL_STENCIL, 0, np_color)
        elif target in [GL.GL_DEPTH_STENCIL, GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT, GL.GL_DEPTH_STENCIL_ATTACHMENT]:
            GL.glClearBufferfv(GL.GL_DEPTH_STENCIL, 0, np_color)



from ..Mesh import Mesh
from ..ColorMap import ColorMap

from glass.utils import checktype
from glass import AttrList

import glm
import numpy as np
import copy

class Surf(Mesh):
    @checktype
    def __init__(self, X:np.ndarray, Y:np.ndarray, Z:np.ndarray,
                 C:np.ndarray=None, back_C:np.ndarray=None,
                 color_map:ColorMap=None, back_color_map:ColorMap=None, 
                 color:(glm.vec3,glm.vec4)=None, back_color:(glm.vec3,glm.vec4)=None,
                 surf_type:Mesh.SurfType=Mesh.SurfType.Smooth, name=""):
        Mesh.__init__(self, name=name, surf_type=surf_type)
        self._XData = X
        self._YData = Y
        self._ZData = Z
        self._CData = C if C is not None else Z

        self._back_CData_user_set = (back_C is not None)
        self._back_CData = back_C if self._back_CData_user_set else self._CData

        Surf._set_colors(self, color, back_color, color_map, back_color_map)
        self.start_building()

    @staticmethod
    def _set_colors(obj, color, back_color, color_map, back_color_map):
        obj._use_color_map = (color is None)
        if color_map is None:
            color_map = ColorMap.parula()

        if color is not None:
            obj.color = color

        if back_color is None:
            back_color = color

        obj._color_map = color_map
        obj._back_color_map_user_set = (back_color_map is not None)
        obj._back_use_color_map = (back_color is None or back_color_map is not None)
        if back_color_map is None:
            back_color_map = copy.deepcopy(color_map)

        if back_color is not None:
            obj.back_color = back_color

        obj._back_color_map = back_color_map

    @staticmethod
    def _build(obj, X, Y, Z, C, back_C):
        obj.should_add_color = False
        rows = X.shape[0]
        cols = X.shape[1]
        color_map = obj._color_map
        back_color_map = obj._back_color_map
        use_color_map = obj._use_color_map
        back_use_color_map = obj._back_use_color_map
        color = obj._color
        back_color = obj._back_color

        xx = X.flatten()
        yy = Y.flatten()
        zz = Z.flatten()
        pos = np.vstack((xx, yy, zz))
        pos = pos.transpose().astype(np.float32)
        it_vertex = np.arange(0, len(xx), dtype=np.uint32)
        jj = it_vertex % cols
        ii = it_vertex // cols
        tt = 1 - ii/(rows - 1)
        ss = jj/(cols - 1)
        tex_coords = np.vstack((ss, tt, np.zeros_like(ss))).transpose().astype(np.float32)
        colors = None
        if use_color_map:
            if len(C.shape) == 2 or C.shape[2] == 1:
                if not color_map.range_user_set:
                    color_map.range = (C.min(), C.max())
                colors = color_map(C).flatten().reshape(len(xx), 4)
            else:
                if C.shape[2] == 4:
                    colors = C.flatten().reshape(len(xx), 4)
                else:
                    colors = np.insert(C, 3, np.ones(C.shape[:2], dtype=np.float32), axis=2).flatten().reshape(len(xx), 4)
        else:
            colors = np.tile([color.r, color.g, color.b, color.a], len(xx)).reshape(len(xx), 4)
        colors = colors.astype(np.float32)

        back_colors = None
        if back_use_color_map:
            if len(back_C.shape) == 2 or back_C.shape[2] == 1:
                if not back_color_map.range_user_set:
                    back_color_map.range = (back_C.min(), back_C.max())
                back_colors = back_color_map(back_C).flatten().reshape(len(xx), 4)
            else:
                if back_C.shape[2] == 4:
                    back_colors = back_C.flatten().reshape(len(xx), 4)
                else:
                    back_colors = np.insert(back_C, 3, np.ones(back_C.shape[:2], dtype=np.float32), axis=2).flatten().reshape(len(xx), 4)
        else:
            back_colors = np.tile([back_color.r, back_color.g, back_color.b, back_color.a], len(xx)).reshape(len(xx), 4)
        back_colors = back_colors.astype(np.float32)

        obj.vertices.reset(
            position=AttrList.fromarray(pos, glm.vec3),
            tangent=AttrList.fromarray(np.zeros_like(pos), glm.vec3),
            bitangent=AttrList.fromarray(np.zeros_like(pos), glm.vec3),
            normal=AttrList.fromarray(np.zeros_like(pos), glm.vec3),
            tex_coord=AttrList.fromarray(tex_coords, glm.vec3),
            color=AttrList.fromarray(colors, glm.vec4),
            back_color=AttrList.fromarray(back_colors, glm.vec4)
        )

        it_vertex = it_vertex[(ii != 0) & (jj != 0)]
        indices0 = it_vertex
        indices1 = it_vertex - 1
        indices2 = it_vertex - 1 - cols
        indices3 = it_vertex
        indices4 = it_vertex - 1 - cols
        indices5 = it_vertex - cols
        indices_mat = np.vstack((indices0, indices1, indices2, indices3, indices4, indices5))
        indices_mat = indices_mat.transpose().reshape(-1, 3)
        obj.indices.reset(indices_mat, dtype=glm.uvec3)

    def build(self):
        X = self._XData
        Y = self._YData
        Z = self._ZData
        C = self._CData if self._CData is not None else Z
        back_C = self._back_CData if self._back_CData is not None else C
        Surf._build(self, X, Y, Z, C, back_C)

    @property
    def XData(self):
        return self._XData
    
    @XData.setter
    @Mesh.param_setter
    def XData(self, X:np.ndarray):
        self._XData = X
    
    @property
    def YData(self):
        return self._YData
    
    @YData.setter
    @Mesh.param_setter
    def YData(self, Y:np.ndarray):
        self._YData = Y
    
    @property
    def ZData(self):
        return self._ZData
    
    @ZData.setter
    @Mesh.param_setter
    def ZData(self, Z:np.ndarray):
        self._ZData = Z
    
    @property
    def CData(self):
        return self._CData
    
    @CData.setter
    @Mesh.param_setter
    def CData(self, C:np.ndarray):
        self._CData = C
        self._use_color_map = True
        if not self._back_CData_user_set:
            self._back_CData = C
            self._back_use_color_map = True

    @property
    def back_CData(self):
        return self._back_CData
    
    @back_CData.setter
    @Mesh.param_setter
    def back_CData(self, C:np.ndarray):
        self._back_CData = C
        self._back_CData_user_set = True
        self._back_use_color_map = True
    
    @property
    def color_map(self):
        return self._color_map
    
    @color_map.setter
    @Mesh.param_setter
    def color_map(self, color_map:ColorMap):
        self._color_map = color_map
        self._use_color_map = True
        if not self._back_color_map_user_set:
            self._back_color_map = color_map
            self._back_use_color_map = True

    @property
    def back_color_map(self):
        return self._back_color_map
    
    @back_color_map.setter
    @Mesh.param_setter
    def back_color_map(self, color_map:ColorMap):
        self._back_color_map = color_map
        self._back_use_color_map = True
        self._back_color_map_user_set = True



from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
import math
from OpenGL import GL

class Circle(Mesh):
    
    def __init__(self, radius:float=1, vertical:bool=False, n_points:int=100,
                 start_angle:float=0, span_angle:float=360,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1),
                 line_width:int=2, name:str=""):
        Mesh.__init__(self, primitive_type=GL.GL_LINE_STRIP, color=color, name=name)
        self.render_hints.line_width = line_width
        self.__radius:float = radius
        self.__vertical:bool = vertical
        self.__n_points:int = n_points
        self.__start_angle:float = start_angle
        self.__span_angle:float = span_angle
        self.start_building()

    def build(self):
        vertices = self.vertices
        radius:float = self.__radius
        vertical:bool = self.__vertical
        n_points:int = self.__n_points
        start_angle:float = self.__start_angle
        span_angle:float = self.__span_angle

        for i in range(n_points):
            theta = (start_angle + i/(n_points-1)*span_angle)/180*math.pi

            position = radius * glm.vec3(math.cos(theta), math.sin(theta), 0)
            bitangent = glm.vec3(-math.sin(theta), math.cos(theta), 0)
            if vertical:
                position = radius * glm.vec3(math.cos(theta), 0, math.sin(theta))
                bitangent = glm.vec3(-math.sin(theta), 0, math.cos(theta))
                
            vertices[i] = Vertex(position=position, bitangent=bitangent, tex_coord=glm.vec3(radius*theta, 0, 0))

        del vertices[n_points:]

    @property
    def line_width(self)->int:
        return self.render_hints.line_width
    
    @line_width.setter
    @checktype
    def line_width(self, line_width:int)->None:
        self.render_hints.line_width = line_width

    @property
    def radius(self)->float:
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float)->None:
        self.__radius = radius

    @property
    def vertical(self)->bool:
        return self.__vertical
    
    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag:bool)->None:
        self.__vertical = flag

    @property
    def n_points(self)->int:
        return self.__n_points
    
    @n_points.setter
    @Mesh.param_setter
    def n_points(self, n_points:int)->None:
        self.__n_points = n_points
    
    @property
    def start_angle(self)->float:
        return self.__start_angle
    
    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, angle:float)->None:
        self.__start_angle = angle

    @property
    def span_angle(self)->float:
        return self.__span_angle
    
    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, angle:float)->None:
        self.__span_angle = angle


from OpenGL import GL
import numpy as np
import re
from ctypes import c_int, pointer

from .GLObject import GLObject
from .GLInfo import GLInfo
from .GlassConfig import GlassConfig
from .GLConfig import GLConfig
from .Uniform import Uniform
from .UniformBlock import UniformBlock
from .ShaderStorageBlock import ShaderStorageBlock
from .ShaderParser import ShaderParser
from .utils import LP_LP_c_char, delete, checktype

_target_type_map = \
{
	"sampler2D": GL.GL_TEXTURE_2D,
	"isampler2D": GL.GL_TEXTURE_2D,
	"usampler2D": GL.GL_TEXTURE_2D,
	"sampler2DMS": GL.GL_TEXTURE_2D_MULTISAMPLE,
	"isampler2DMS": GL.GL_TEXTURE_2D_MULTISAMPLE,
	"usampler2DMS": GL.GL_TEXTURE_2D_MULTISAMPLE,

	"sampler2DArray": GL.GL_TEXTURE_2D_ARRAY,
	"isampler2DArray": GL.GL_TEXTURE_2D_ARRAY,
	"usampler2DArray": GL.GL_TEXTURE_2D_ARRAY,
	"sampler2DMSArray": GL.GL_TEXTURE_2D_MULTISAMPLE_ARRAY,
	"isampler2DMSArray": GL.GL_TEXTURE_2D_MULTISAMPLE_ARRAY,
	"usampler2DMSArray": GL.GL_TEXTURE_2D_MULTISAMPLE_ARRAY,

	"samplerCube": GL.GL_TEXTURE_CUBE_MAP,
	"samplerCubeArray": GL.GL_TEXTURE_CUBE_MAP_ARRAY,

	"image2D": GL.GL_TEXTURE_2D,
	"iimage2D": GL.GL_TEXTURE_2D,
	"uimage2D": GL.GL_TEXTURE_2D
}

class CompileError(Exception):
	pass

class CompileWarning(Warning):
	pass

class LinkError(Exception):
	pass

class LinkWarning(Warning):
	pass

class GPUProgram(GLObject):

	_basic_info = \
	{
		"gen_func": GL.glCreateProgram,
		"bind_func": GL.glUseProgram,
		"del_func": GL.glDeleteProgram,
		"target_type": None,
		"binding_type": GL.GL_CURRENT_PROGRAM,
		"need_number": False,
	}

	__message_prefix1 = re.compile(r"(?P<shader_type>\w+) info\n-{4,}\n0?\((?P<line_number>\d+)\) : (?P<message_type>\w+) ")
	__message_prefix2 = re.compile(r"(?P<shader_type>\w+) info\n-{4,}\n(?P<message_type>\w+): ")
	__message_prefix3 = re.compile(r"\n0?\((?P<line_number>\d+)\) : (?P<message_type>\w+) ")
	__active_program = {}

	def __init__(self):
		GLObject.__init__(self)

		self._attributes_info = {}
		self._acceptable_primitives = []
		self._uniforms_info = {}
		self._uniform_blocks_info = {}
		self._shader_storage_blocks_info = {}
		self._structs_info = {}
		self._outs_info = {}

		self._sampler_map = {}
		self._uniform_map = {}
		self._uniform_block_map = {}
		self._shader_storage_block_map = {}

		self._uniform = Uniform(self)
		self._uniform_block = UniformBlock(self)
		self._shader_storage_block = ShaderStorageBlock(self)

		self._is_linked = False
		self._linked_but_not_applied = False
		self._uniform_not_set_warning = True

	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return (id(self) == id(other))

	@delete
	def bind(self):
		pass
	
	@delete
	def unbind(self):
		pass
	
	@delete
	def is_bound(self):
		pass

	def __getitem__(self, name:str):
		if not self._is_linked:
			self._link()

		if name in self._uniform_map:
			return self._uniform[name]
		elif name in self._shader_storage_block_map:
			return self._shader_storage_block[name]
		elif name in self._uniform_block_map:
			return self._uniform_block[name]
		else:
			error_message = "'" + name + "' is not defined in following files:\n"
			error_message += "\n".join(self.related_files)
			raise NameError(error_message)

	def __setitem__(self, name:str, value):
		if not self._is_linked:
			self._link()

		if name in self._uniform_map:
			self._uniform[name] = value
		elif name in self._shader_storage_block_map:
			self._shader_storage_block[name] = value
		elif name in self._uniform_block_map:
			self._uniform_block[name] = value
		else:
			error_message = "'" + name + "' is not defined in following files:\n"
			error_message += "\n".join(self.related_files)
			raise NameError(error_message)
		
	def __contains__(self, name:str):
		if not self._is_linked:
			self._link()

		return (name in self._uniform_map or \
			    name in self.buffer._blocks_info or \
			    name in self.uniform_block._blocks_info)

	@property
	def uniform(self):
		if not self._is_linked:
			self._link()
			
		return self._uniform

	@property
	def uniform_block(self):
		if not self._is_linked:
			self._link()
			
		return self._uniform_block

	@property
	def buffer(self):
		if not self._is_linked:
			self._link()

		return self._shader_storage_block
	
	def download(self, var):
		id_var = id(var)
		if id_var not in ShaderStorageBlock._bound_vars:
			raise ValueError(f"{var} is not bound with any shader storage blocks")

		for block_var in self._shader_storage_block._block_var_map.values():
			if block_var._bound_var is var:
				for ssbo in ShaderStorageBlock._bound_vars[id_var].values():
					if block_var in ssbo._bound_block_vars:
						ssbo.download()

	def compile(self, shader_type, file_name=None):
		pass

	def _apply(self):
		pass

	def _link(self):
		pass

	@property
	def is_linked(self):
		return self._is_linked

	def use(self):
		if not self._is_linked:
			self._link()
		if self._linked_but_not_applied:
			self._apply()

		current_context = GLConfig.buffered_current_context
		if current_context not in GPUProgram.__active_program or \
		   GPUProgram.__active_program[current_context] != self._id:
			GL.glUseProgram(self._id)
			GPUProgram.__active_program[current_context] = self._id

	def stop_using(self):
		if not self.is_using:
			return

		GL.glUseProgram(0)

	@property
	def is_using(self):
		return (self._id != 0 and GPUProgram.active_id == self._id)
	
	@property
	def uniform_not_set_warning(self):
		return self._uniform_not_set_warning
	
	@uniform_not_set_warning.setter
	@checktype
	def uniform_not_set_warning(self, flag:bool):
		self._uniform_not_set_warning = flag

	@property
	def related_files(self):
		return []
	
	def _format_error_warning(self, message):
		used_shaders = []
		def _replace_message1(match):
			message_type = match.group("message_type")
			shader_type = match.group("shader_type")
			line_number = int(match.group('line_number'))

			used_shader = None
			if shader_type == "Vertex":
				used_shader = self.vertex_shader
			elif shader_type == "Fragment":
				used_shader = self.fragment_shader
			elif shader_type == "Geometry":
				used_shader = self.geometry_shader
			elif shader_type == "Tessellation control":
				used_shader = self.tess_ctrl_shader
			elif shader_type == "Tessellation evaluation":
				used_shader = self.tess_eval_shader
			
			if used_shader is not None:
				used_shaders.append(used_shader)
				return used_shader._line_message_map[line_number].replace("\\", "/").replace("./", "").format(message_type=message_type)
			else:
				return message_type + ": "
		
		def _replace_message2(match):
			message_type = match.group("message_type")
			return message_type + ": "
		
		def _replace_message3(match):
			message_type = match.group("message_type")
			line_number = int(match.group('line_number'))

			used_shader = None
			if used_shaders:
				used_shader = used_shaders[-1]
			
			if used_shader is not None:
				return "\n" + used_shader._line_message_map[line_number].replace("\\", "/").replace("./", "").format(message_type=message_type)

		message = GPUProgram.__message_prefix1.sub(_replace_message1, message)
		message = GPUProgram.__message_prefix2.sub(_replace_message2, message)
		message = GPUProgram.__message_prefix3.sub(_replace_message3, message)
		lines = message.split("\n")
		warning_messages = []
		error_messages = []
		last = "error"
		for line in lines:
			line = line.strip(" \t\n\r")
			if not line:
				continue

			line = line.replace("no program defined", "no 'void main()' defined")

			if "error" in line.lower():
				error_messages.append(line)
				last = "error"
			elif "warning" in line.lower():
				warning_messages.append(line)
				last = "warning"
			elif last == "error":
				error_messages.append(line)
			elif last == "warning":
				warning_messages.append(line)

		return error_messages, warning_messages

	def _resolve_one_uniform(self, var, uniform_map):
		var_type = var["type"]
		var_name = var["name"]
		var_subscript_chain = var["subscript_chain"]
		if var_type in GLInfo.atom_type_names:
			var_atoms = [var]
			var["atoms"] = var_atoms
			# var["location"] = GL.glGetUniformLocation(self._id, var["name"])
			uniform_map[var_name] = var
			return var_atoms

		if "[" in var_type:
			pos_start = var_type.find("[")
			pos_end = var_type.find("]", pos_start)
			base_type = var_type[:pos_start]
			atom_type = base_type + var_type[pos_end+1:]
			
			var_atoms = []
			if pos_end - pos_start > 1:
				num = int(var_type[pos_start+1 : pos_end])
				for i in range(num):
					atom_info = \
					{
						"name": var_name + "[" + str(i) + "]",
						"type": atom_type,
						"subscript_chain": [*var_subscript_chain, ("getitem", i)]
					}
					var_atoms.extend( self._resolve_one_uniform(atom_info, self._uniform_map) )
			else:
				atom_info = \
				{
					"name": var_name + "[{0}]",
					"type": atom_type,
					"subscript_chain": [*var_subscript_chain, ("getitem", "{0}")]
				}
				var_atoms.extend( self._resolve_one_uniform(atom_info, self._uniform_map) )
			var["atoms"] = var_atoms
			uniform_map[var_name] = var

			return var_atoms

		if GlassConfig.debug and var_type not in self._structs_info:
			raise TypeError("type " + var_type + " is not defined in shader program")

		var_atoms = []
		for member in self._structs_info[var_type]["members"].values():
			member_name = member["name"]
			uniform_info = \
			{
				"name": var_name + "." + member_name,
				"type": member["type"],
				"subscript_chain": [*var_subscript_chain, ("getattr", member_name)]
			}
			var_atoms.extend( self._resolve_one_uniform(uniform_info, self._uniform_map) )
		var["atoms"] = var_atoms
		uniform_map[var_name] = var

		return var_atoms

	def _resolve_uniforms(self):
		for uniform_info in self._uniforms_info.values():
			self._resolve_one_uniform(uniform_info, self._uniform_map)

		for uniform in self._uniform_map.values():
			for atom in uniform["atoms"]:
				atom_type = atom["type"]
				atom_name = atom["name"]
				if "sampler" in atom_type or "image" in atom_type:
					self._sampler_map[atom_name] = \
					{
						"location": -1,
						"sampler": None,
						"target_type": _target_type_map[atom_type]
					}

	def _resolve_one_uniform_block(self, block_info):
		uniform_block = {}
		block_name = block_info["name"]
		block_var_name = block_info["var_name"]
		uniform_block["name"] = block_name
		uniform_block["atoms"] = []

		for member in block_info["members"].values():
			member_name = member["name"]
			if block_var_name:
				member_name = block_var_name + "." + member_name
			uniform_info = \
			{
				"name": member_name,
				"type": member["type"],
				"subscript_chain": [("getattr", member_name)]
			}
			uniform_block["atoms"].extend( self._resolve_one_uniform(uniform_info, self._uniform_block_map) )

		self._uniform_block_map[block_name] = uniform_block
		return uniform_block["atoms"]

	def _apply_uniform_blocks(self):
		for block_name, block_info in self._uniform_blocks_info.items():
			atoms = self._uniform_block_map[block_name]["atoms"]
			len_atoms = len(atoms)
			atom_names = []
			for atom in atoms:
				atom_names.append(atom["name"])

			block_index = GL.glGetUniformBlockIndex(self._id, block_name)
			len_var_name = len(block_info["var_name"])

			block_size = np.array([0], dtype=int)
			GL.glGetActiveUniformBlockiv(self._id, block_index, GL.GL_UNIFORM_BLOCK_DATA_SIZE, block_size)
			block_size = block_size[0]
			block_info["size"] = block_size
			block_info["index"] = block_index
			
			atom_indices = np.array([0]*len_atoms, dtype=np.uint)
			GL.glGetUniformIndices(self._id, len_atoms, LP_LP_c_char(atom_names), atom_indices)
			
			atom_offsets = np.array([0]*len_atoms, dtype=int)
			GL.glGetActiveUniformsiv(self._id, len_atoms, atom_indices, GL.GL_UNIFORM_OFFSET, atom_offsets)

			array_strides = np.array([0]*len_atoms, dtype=int)
			GL.glGetActiveUniformsiv(self._id, len_atoms, atom_indices, GL.GL_UNIFORM_ARRAY_STRIDE, array_strides)

			block_members = block_info["members"]
			for atom, array_stride, atom_offset in zip(atoms, array_strides, atom_offsets):
				atom_name = atom["name"]
				member_name = ShaderParser.array_basename(atom_name)
				if len_var_name > 0:
					member_name = member_name[len_var_name+1:]

				stride = int(array_stride)
				member_type = block_members[member_name]["type"]
				offset = atom_offset + stride * ShaderParser.index_offset(member_type, atom_name)
				atom["offset"] = int(offset)
				atom["stride"] = stride

	def _resolve_uniform_blocks(self):
		for block_info in self._uniform_blocks_info.values():
			self._resolve_one_uniform_block(block_info)
		
		blocks_with_var_name = {}

		for block_name, block_info in self._uniform_blocks_info.items():
			var_name = block_info["var_name"]
			if var_name:
				blocks_with_var_name[var_name] = block_name

		for var_name, block_name in blocks_with_var_name.items():
			self._uniform_blocks_info[var_name] = self._uniform_blocks_info[block_name]
			self._uniform_block_map[var_name] = self._uniform_block_map[block_name]

		# self._apply_uniform_blocks()

	def _resolve_one_shader_storage_block(self, block_info):
		shader_storage_block = {}
		block_name = block_info["name"]
		block_var_name = block_info["var_name"]

		shader_storage_block["name"] = block_name
		shader_storage_block["atoms"] = []

		for member in block_info["members"].values():
			member_name = member["name"]
			if block_var_name:
				member_name = block_var_name + "." + member_name
			uniform_info = \
			{
				"name": member_name,
				"type": member["type"],
				"subscript_chain": [("getattr", member_name)]
			}
			shader_storage_block["atoms"].extend( self._resolve_one_uniform(uniform_info, self._shader_storage_block_map) )

		self._shader_storage_block_map[block_name] = shader_storage_block
		if block_var_name:
			self._shader_storage_block_map[block_var_name] = shader_storage_block

		return shader_storage_block["atoms"]

	def _apply_shader_storage_blocks(self):
		for block_name, block_info in self._shader_storage_blocks_info.items():
			atoms = self._shader_storage_block_map[block_name]["atoms"]

			block_index = GL.glGetProgramResourceIndex(self._id, GL.GL_SHADER_STORAGE_BLOCK, block_name)
			block_info["index"] = block_index

			block_var_name = block_info["var_name"]
			len_var_name = len(block_var_name)
			if block_var_name:
				raise ValueError("Shader Storage Block with variable name is not supported")

			props = np.array([GL.GL_OFFSET, GL.GL_ARRAY_STRIDE, GL.GL_TOP_LEVEL_ARRAY_STRIDE])
			length = c_int()
			block_members = block_info["members"]
			for atom in atoms:
				atom_offset_stride = np.array([0,0,0], dtype=np.int32)
				atom_name = re.sub(r"\[\d+\]", "[0]", atom["name"].format(0))
				atom_index = GL.glGetProgramResourceIndex(self._id, GL.GL_BUFFER_VARIABLE, atom_name)
				GL.glGetProgramResourceiv(self._id, GL.GL_BUFFER_VARIABLE, atom_index, 3, props, atom_offset_stride.nbytes, pointer(length), atom_offset_stride)
				atom_offset_stride = [int(x) for x in atom_offset_stride]

				member_name = ShaderParser.array_basename(atom_name)
				# member_name = ShaderParser.array_basename(atom["name"])
				if block_var_name:
					member_name = member_name[len_var_name+1:]
				# atom_offset = atom_offset_stride[0] + atom_offset_stride[1] * ShaderParser.index_offset(block_members[member_name]["type"], atom["name"].format(0))
				atom_offset = atom_offset_stride[0] + atom_offset_stride[1] * ShaderParser.index_offset(block_members[member_name]["type"], atom_name)
				atom_stride = min(atom_offset_stride[1], atom_offset_stride[2])
				if atom_stride == 0:
					atom_stride = max(atom_offset_stride[1], atom_offset_stride[2])

				atom["offset"] = int(atom_offset)
				atom["stride"] = int(atom_stride)

	def _resolve_shader_storage_blocks(self):
		for block_info in self._shader_storage_blocks_info.values():
			self._resolve_one_shader_storage_block(block_info)

		blocks_with_var_name = {}
		for block_name, block_info in self._shader_storage_blocks_info.items():
			var_name = block_info["var_name"]
			if var_name:
				blocks_with_var_name[var_name] = block_name

		for var_name in blocks_with_var_name:
			self._shader_storage_blocks_info[var_name] = block_info
			self._shader_storage_block_map[var_name] = self._shader_storage_block_map[block_name]

		# self._apply_shader_storage_blocks()


from .Light import Light, FlatLight
from glass.DictList import DictList
from glass.ShaderStorageBlock import ShaderStorageBlock

import glm

class DirLight(Light):
    pass

class FlatDirLight(FlatLight):

    def __init__(self, dir_light:DirLight):
        self.direction = glm.vec3(0, 1, 0)
        self.abs_orientation = glm.quat(1, 0, 0, 0)
        self.max_back_offset = 0
        FlatLight.__init__(self, dir_light)

class DirLights(ShaderStorageBlock.HostClass):

    def __init__(self):
        ShaderStorageBlock.HostClass.__init__(self)
        self.dir_lights = DictList()

    def __getitem__(self, key:(str,int)):
        return self.dir_lights[key]
    
    @ShaderStorageBlock.HostClass.not_const
    def __setitem__(self, key:(str,int), dir_light:FlatDirLight):
        self.dir_lights[key] = dir_light

    @ShaderStorageBlock.HostClass.not_const
    def __delitem__(self, key:(str,int)):
        del self.dir_lights[key]

    def __contains__(self, path_str:str):
        return (path_str in self.dir_lights)
    
    def __len__(self):
        return len(self.dir_lights)
    
    def __iter__(self):
        return iter(self.dir_lights)
    
    def keys(self):
        return self.dir_lights.keys()

    @property
    def n_dir_lights(self):
        return len(self.dir_lights)


from .Manipulator import Manipulator

from glass.RenderHints import RenderHints
from glass.utils import checktype

from OpenGL import GL
import math
import glm
import datetime

class ModelViewManipulator(Manipulator):

    def __init__(self, distance:float=5, azimuth:float=0, elevation:float=0):
        Manipulator.__init__(self)
        
        self.__is_left_pressed = False
        self.__left_press_global_posF = glm.vec2(0, 0)
        self.__left_press_azimuth = 0
        self.__left_press_elevation = 0

        self.__is_right_pressed = False
        self.__right_press_global_posF = glm.vec2(0, 0)
        self.__right_press_offset = glm.vec2(0, 0)

        self.__distance = distance
        self.__azimuth = azimuth
        self.__elevation = elevation
        self.__offset = glm.vec2(0, 0)
        self.__sensitivity = 1

    @property
    def distance(self)->float:
        return self.__distance
    
    @distance.setter
    @checktype
    def distance(self, distance:float):
        self.__distance = distance
        self.__update_camera()
        self.camera.screen.update()

    @property
    def azimuth(self):
        return self.__azimuth
    
    @azimuth.setter
    @checktype
    def azimuth(self, azimuth:float):
        self.__azimuth = azimuth
        self.__update_camera()
        self.camera.screen.update()

    @property
    def elevation(self):
        return self.__elevation
    
    @elevation.setter
    @checktype
    def elevation(self, elevation:float):
        self.__elevation = elevation
        self.__update_camera()
        self.camera.screen.update()

    @property
    def sensitivity(self)->float:
        return self.__sensitivity
    
    @sensitivity.setter
    @checktype
    def sensitivity(self, sensitivity:float):
        self.__sensitivity = sensitivity

    def __update_camera(self):
        azimuth = self.__azimuth/180*math.pi
        elevation = self.__elevation/180*math.pi

        self.camera.position.x = self.__distance*math.cos(elevation)*math.sin(azimuth)
        self.camera.position.y = -self.__distance*math.cos(elevation)*math.cos(azimuth)
        self.camera.position.z = self.__distance*math.sin(elevation)

        if glm.length(self.__offset) > 1E-6:
            right = glm.vec3(math.cos(azimuth), math.sin(azimuth), 0)
            forward = glm.vec3(-math.sin(elevation) * math.sin(azimuth), math.sin(elevation) * math.cos(azimuth), math.cos(elevation))
            self.camera.position += (self.__offset.x * right + self.__offset.y * forward)

        self.camera.pitch = -self.__elevation
        self.camera.yaw = self.__azimuth

    def startup(self):
        self.__update_camera()

    def on_mouse_pressed(self, button:Manipulator.MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2):
        if button == Manipulator.MouseButton.LeftButton:
            self.__is_left_pressed = True
            self.__left_press_global_posF = global_pos
            self.__left_press_azimuth = self.__azimuth
            self.__left_press_elevation = self.__elevation
        elif button == Manipulator.MouseButton.RightButton:
            self.__is_right_pressed = True
            self.__right_press_global_posF = global_pos
            self.__right_press_offset = self.__offset
        elif button == Manipulator.MouseButton.XButton1:
            self.__sensitivity /= pow(2, 1/2)
        elif button == Manipulator.MouseButton.XButton2:
            self.__sensitivity *= pow(2, 1/2)

        return False

    def on_mouse_released(self, button:Manipulator.MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2):
        if button == Manipulator.MouseButton.LeftButton:
            self.__is_left_pressed = False
        elif button == Manipulator.MouseButton.RightButton:
            self.__is_right_pressed = False

        return False
    
    def on_mouse_double_clicked(self, button:Manipulator.MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        if button == Manipulator.MouseButton.LeftButton:
            self.__offset = glm.vec2(0, 0)
            self.__update_camera()
            return True
        
        return False

    def on_mouse_moved(self, screen_pos:glm.vec2, global_pos:glm.vec2):
        if self.__is_left_pressed:
            d = global_pos - self.__left_press_global_posF
            dx = d.x
            dy = d.y

            d_pitch = dy / self.camera.screen.height() * 200 * self.sensitivity
            d_yaw = -dx / self.camera.screen.width() * 200 * self.sensitivity
            
            self.__azimuth = self.__left_press_azimuth + d_yaw
            self.__elevation = self.__left_press_elevation + d_pitch
            self.__update_camera()

            return True
        
        elif self.__is_right_pressed:
            d = global_pos - self.__right_press_global_posF
            dx = -d.x
            dy = d.y

            offset_delta = glm.vec2()
            offset_delta.x = dx / self.camera.screen.width() * 2 * self.__distance * self.camera.tan_half_fov * self.camera.aspect
            offset_delta.y = dy / self.camera.screen.height() * 2 * self.__distance * self.camera.tan_half_fov
            self.__offset = self.__right_press_offset + offset_delta
            self.__update_camera()

            return True

        return False

    def on_wheel_scrolled(self, angle:glm.vec2, screen_pos:glm.vec2, global_pos:glm.vec2):
        if self.__is_left_pressed or self.__is_right_pressed:
            return False
        
        n = angle.y/120
        scale = pow(2, n/6)
        if self.camera.projection_mode.value == 0:
            self.camera.fov /= scale
        else:
            self.camera.height /= scale

        self.__update_camera()
        return True

    def on_key_pressed(self, key:Manipulator.Key)->bool:
        if key == Manipulator.Key.Key_R:
            polygon_mode = self.camera.screen.render_hints.polygon_mode
            if polygon_mode in [GL.GL_FILL, RenderHints.inherit]:
                self.camera.screen.render_hints.polygon_mode = GL.GL_LINE
                self.camera.screen.render_hints.line_width = 1
            elif polygon_mode == GL.GL_LINE:
                self.camera.screen.render_hints.polygon_mode = GL.GL_POINT
                self.camera.screen.render_hints.point_size = 1.5
            elif polygon_mode == GL.GL_POINT:
                self.camera.screen.render_hints.polygon_mode = GL.GL_FILL
            return True
        elif key == Manipulator.Key.Key_P:
            now = datetime.datetime.now()
            file_name = "capture_" + now.strftime("%Y-%m-%d_%H-%M-%S") + ".png"
            self.camera.screen.capture(file_name)
        
    def on_key_repeated(self, keys)->bool:
        should_update = False
        if Manipulator.Key.Key_W in keys and \
           Manipulator.Key.Key_S not in keys:
            self.__distance /= pow(2, 1/self.camera.screen.smooth_fps)
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_S in keys and \
             Manipulator.Key.Key_W not in keys:
            self.__distance *= pow(2, 1/self.camera.screen.smooth_fps)
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_A in keys and \
             Manipulator.Key.Key_D not in keys:
            self.__offset.x -= 0.5/self.camera.screen.smooth_fps * self.__distance * self.camera.tan_half_fov
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_D in keys and \
             Manipulator.Key.Key_A not in keys:
            self.__offset.x += 0.5/self.camera.screen.smooth_fps * self.__distance * self.camera.tan_half_fov
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_E in keys and \
             Manipulator.Key.Key_C not in keys:
            self.__offset.y += 0.5/self.camera.screen.smooth_fps * self.__distance * self.camera.tan_half_fov
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_C in keys and \
             Manipulator.Key.Key_E not in keys:
            self.__offset.y -= 0.5/self.camera.screen.smooth_fps * self.__distance * self.camera.tan_half_fov
            self.__update_camera()
            should_update = True
        if Manipulator.Key.Key_F in keys:
            screen = self.camera.screen
            print("fps:", screen.fps, "draw calls:", screen.draw_calls)
            should_update = True

        return should_update


from .WeakRef import WeakRef

class WeakSet:

    class iterator:
        def __init__(self, _iterator):
            self._iterator = _iterator

        def __next__(self):
            value = self._iterator.__next__()
            return value.value
        
        def __iter__(self):
            return self

    def __init__(self):
        self._set = set()

    def __iter__(self):
        result = WeakSet.iterator(self._set.__iter__())
        return result
    
    def __len__(self):
        return len(self._set)
    
    def __bool__(self):
        return bool(self._set)
    
    def __contains__(self, value):
        return (WeakRef(value) in self._set)
    
    def __repr__(self):
        return self._set.__repr__()
    
    def __sub__(self, other):
        result = WeakSet()
        if isinstance(other, WeakSet):
            result._set = self._set - other._set
        else:
            result._set = self._set - map(lambda x:WeakRef(x), other)

        return result
    
    def __add__(self, other):
        result = WeakSet()
        if isinstance(other, WeakSet):
            result._set = self._set + other._set
        else:
            result._set = self._set + map(lambda x:WeakRef(x), other)

        return result
    
    def __isub__(self, other):
        result = WeakSet()
        if isinstance(other, WeakSet):
            self._set -= other._set
        else:
            self._set -= map(lambda x:WeakRef(x), other)

        return result
    
    def __iadd__(self, other):
        result = WeakSet()
        if isinstance(other, WeakSet):
            self._set += other._set
        else:
            self._set += map(lambda x:WeakRef(x), other)

        return result

    def add(self, value):
        self._set.add(WeakRef(value))

    def remove(self, value):
        ref_value = WeakRef(value)
        if ref_value in self._set:
            self._set.remove(ref_value)

    def pop(self):
        return self._set.pop().value

    def clear(self):
        self._set.clear()
    
    def update(self, new_set):
        if isinstance(new_set, WeakSet):
            self._set.update(new_set._set)
        else:
            self._set.update(map(lambda x:WeakRef(x), new_set))



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class SphericalCapTop(Mesh):

    @checktype
    def __init__(self, base_radius:float=1, height:float=0.5,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 n_lon_divide:int=100, start_lon:float=0, span_lon:float=360,
                 n_lat_divide:int=50, name:str="", block=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__base_radius = base_radius
        self.__height = height
        self.__start_lon = start_lon
        self.__span_lon = span_lon
        self.__n_lon_divide = n_lon_divide
        self.__n_lat_divide = n_lat_divide
        self.start_building()

    def build(self):
        self.is_closed = True

        vertices = self.vertices
        indices = self.indices
        base_radius = self.__base_radius
        height = self.__height
        radius = (base_radius**2+height**2) / (2*height)
        start_lon = self.__start_lon/180*math.pi
        span_lon = self.__span_lon/180*math.pi
        start_lat = math.asin((base_radius**2-height**2)/(base_radius**2+height**2))
        span_lat = math.pi/2 - start_lat
        n_lon_divide = self.__n_lon_divide
        n_lat_divide = self.__n_lat_divide
        h_delta = (base_radius**2-height**2)/(2*height)

        i_vertex = 0
        i_index = 0

        for i in range(n_lon_divide):
            theta = start_lon + span_lon*i/(n_lon_divide - 1)
            s = theta/(2*math.pi)
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)

            for j in range(n_lat_divide):
                phi = start_lat + span_lat*j/(n_lat_divide-1)
                t = phi/math.pi + 0.5
                cos_phi = math.cos(phi)
                sin_phi = math.sin(phi)

                vertex = Vertex()

                vertex.tangent = 2*math.pi*radius*cos_phi*glm.vec3(-sin_theta, cos_theta, 0)
                vertex.bitangent = math.pi*radius*glm.vec3(-sin_phi*cos_theta, -sin_phi*sin_theta, cos_phi)
                vertex.normal = glm.vec3(cos_phi*cos_theta, cos_phi*sin_theta, sin_phi)
                vertex.position = radius * vertex.normal
                vertex.position.z -= h_delta
                vertex.tex_coord = glm.vec3(s, t, 0)

                vertices[i_vertex] = vertex
                i_vertex += 1

                if i > 0 and j > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex-1
                    triangle[1] = i_vertex-1-n_lat_divide
                    triangle[2] = i_vertex-1-n_lat_divide-1
                    indices[i_index] = triangle
                    i_index += 1

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex-1
                    triangle[1] = i_vertex-1-n_lat_divide-1
                    triangle[2] = i_vertex-1-1
                    indices[i_index] = triangle
                    i_index += 1

                    yield
        
        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_lat_divide(self):
        return self.__n_lat_divide
    
    @n_lat_divide.setter
    @Mesh.param_setter
    def n_lat_divide(self, n:int):
        self.__n_lat_divide = n

    @property
    def n_lon_divide(self):
        return self.__n_lon_divide
    
    @n_lon_divide.setter
    @Mesh.param_setter
    def n_lon_divide(self, n:int):
        self.__n_lon_divide = n

    @property
    def start_lon(self):
        return self.__start_lon
    
    @start_lon.setter
    @Mesh.param_setter
    def start_lon(self, angle:float):
        self.__start_lon = angle

    @property
    def span_lon(self):
        return self.__span_lon
    
    @span_lon.setter
    @Mesh.param_setter
    def span_lon(self, angle:float):
        self.__span_lon = angle

    @property
    def base_radius(self):
        return self.__base_radius
    
    @base_radius.setter
    @Mesh.param_setter
    def base_radius(self, base_radius:float):
        self.__base_radius = base_radius

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height
        


from .GLConfig import GLConfig

class ImageUnits:
    def __init__(self):
        self._unit_image_map = {}
        self._image_unit_map = {}

    def __getitem__(self, unit:int)->tuple:
        key = (GLConfig.buffered_current_context, unit)
        if key not in self._unit_image_map:
            return None
        
        return self._unit_image_map[key]
    
    def __setitem__(self, unit:int, image:tuple):
        context = GLConfig.buffered_current_context

        unit_key = (context, unit)
        if unit_key in self._unit_image_map:
            old_image = self._unit_image_map[unit_key]
            if old_image != image:
                del self._image_unit_map[context, old_image]

        image_key = (context, image)
        if image_key in self._image_unit_map:
            old_unit = self._image_unit_map[image_key]
            if old_unit != unit:
                del self._unit_image_map[context, old_unit]

        self._unit_image_map[unit_key] = image
        self._image_unit_map[image_key] = unit

    def unit_of_image(self, image:tuple)->int:
        key = (GLConfig.buffered_current_context, image)
        if key not in self._image_unit_map:
            return None
        
        return self._image_unit_map[key]
    
    @property
    def available_unit(self):
        context = GLConfig.buffered_current_context

        for i in range(GLConfig.max_image_units):
            key = (context, i)
            if key not in self._unit_image_map:
                return i
            
            if self._unit_image_map[key] == 0:
                return i
            
        return None

ImageUnits = ImageUnits()
